from django.db import models

from clickhouse_backend.models import fields
from clickhouse_backend.utils.timezone import get_timezone

from .base import Func

__all__ = [
    "toStartOfMinute",
    "toStartOfFiveMinutes",
    "toStartOfTenMinutes",
    "toStartOfFifteenMinutes",
    "toStartOfHour",
    "toYYYYMM",
    "toYYYYMMDD",
    "toYYYYMMDDhhmmss",
    "toYearWeek",
]


class toYYYYMM(Func):
    output_field = fields.UInt32Field()

    def __init__(self, *expressions):
        arity = len(expressions)
        if arity < 1 or arity > 2:
            raise TypeError(
                "'%s' takes 1 or 2 arguments (%s given)"
                % (
                    self.__class__.__name__,
                    len(expressions),
                )
            )
        if arity == 2 and isinstance(expressions[1], str):
            expressions = (expressions[0], models.Value(expressions[1]))
        else:
            expressions = (expressions[0], models.Value(get_timezone()))

        super().__init__(*expressions)


class toYYYYMMDD(toYYYYMM):
    pass


class toYYYYMMDDhhmmss(toYYYYMM):
    output_field = fields.UInt64Field()


class toStartOfMinute(Func):
    output_field = models.fields.DateTimeField()

    def __init__(self, *expressions):
        arity = len(expressions)
        if arity < 1 or arity > 1:
            raise TypeError(
                "'%s' takes 1 argument (%s given)"
                % (
                    self.__class__.__name__,
                    len(expressions),
                )
            )

        super().__init__(*expressions)


class toStartOfFiveMinutes(toStartOfMinute):
    pass


class toStartOfTenMinutes(toStartOfMinute):
    pass


class toStartOfFifteenMinutes(toStartOfMinute):
    pass


class toStartOfHour(toStartOfMinute):
    pass


class toYearWeek(Func):
    output_field = fields.UInt32Field()

    def __init__(self, *expressions):
        arity = len(expressions)
        if arity < 1 or arity > 3:
            raise TypeError(
                "'%s' takes between 1 and 3 arguments (%s given)"
                % (
                    self.__class__.__name__,
                    len(expressions),
                )
            )

        expressions = (
            expressions[0],
            *(models.Value(expr) for expr in expressions[1:]),
        )

        super().__init__(*expressions)



from django.db.models.expressions import ExpressionList, F, Func
from django.db.models.indexes import IndexExpression, names_digest, split_identifier
from django.db.models.sql import Query

from clickhouse_backend.models.engines import BaseMergeTree

__all__ = [
    "Index",
    "MinMax",
    "Set",
    "NgrambfV1",
    "TokenbfV1",
    "BloomFilter",
]


class Index:
    suffix = "idx"

    max_name_length = 30

    def __init__(
        self,
        *expressions,
        fields=(),
        name=None,
        type=None,
        granularity=None,
    ):
        if not isinstance(type, Type):
            raise ValueError("Index type must be a Type instance.")
        if not isinstance(granularity, int) and granularity > 0:
            raise ValueError("Index granularity must be a positive integer.")
        if not isinstance(fields, (list, tuple)):
            raise ValueError("Index.fields must be a list or tuple.")
        if not expressions and not fields:
            raise ValueError(
                "At least one field or expression is required to define an index."
            )
        if expressions and fields:
            raise ValueError(
                "Index.fields and expressions are mutually exclusive.",
            )
        if not name:
            raise ValueError("An index must be named.")
        if fields and not all(isinstance(field, str) for field in fields):
            raise ValueError("Index.fields must contain only strings with field names.")
        self.fields = list(fields)
        # A list of 2-tuple with the field name and ordering ("" or "DESC").
        self.fields_orders = [
            (field_name[1:], "DESC") if field_name.startswith("-") else (field_name, "")
            for field_name in self.fields
        ]
        self.include = []
        self.condition = None
        self.name = name or ""
        self.type = type
        self.granularity = granularity
        self.expressions = tuple(
            F(expression) if isinstance(expression, str) else expression
            for expression in expressions
        )

    @property
    def contains_expressions(self):
        return bool(self.expressions)

    def index_sql(self, model, schema_editor):
        return self.create_sql(model, schema_editor, inline=True)

    def create_sql(self, model, schema_editor, **kwargs):
        engine = getattr(model._meta, "engine", None)
        if engine and not isinstance(engine, BaseMergeTree):
            raise TypeError(
                "Only MergeTree family support indexes. Refer https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/mergetree#table_engine-mergetree-data_skipping-indexes"
            )
        if self.expressions:
            index_expressions = []
            for expression in self.expressions:
                index_expression = IndexExpression(expression)
                index_expression.set_wrapper_classes(schema_editor.connection)
                index_expressions.append(index_expression)
            expressions = ExpressionList(*index_expressions).resolve_expression(
                Query(model, alias_cols=False),
            )
            fields = None
            col_suffixes = None
        else:
            fields = [
                model._meta.get_field(field_name)
                for field_name, _ in self.fields_orders
            ]
            col_suffixes = [order[1] for order in self.fields_orders]
            expressions = None
        return schema_editor._create_index_sql(
            model,
            fields=fields,
            name=self.name,
            col_suffixes=col_suffixes,
            type=self.type,
            granularity=self.granularity,
            expressions=expressions,
            **kwargs,
        )

    def remove_sql(self, model, schema_editor, **kwargs):
        engine = getattr(model._meta, "engine", None)
        if engine and not isinstance(engine, BaseMergeTree):
            raise TypeError(
                "Only MergeTree family support indexes. Refer https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/mergetree#table_engine-mergetree-data_skipping-indexes"
            )
        return schema_editor._delete_index_sql(model, self.name, **kwargs)

    def deconstruct(self):
        path = "%s.%s" % (self.__class__.__module__, self.__class__.__name__)
        if path.startswith("clickhouse_backend.models.indexes"):
            path = path.replace(
                "clickhouse_backend.models.indexes", "clickhouse_backend.models"
            )
        kwargs = {"name": self.name}
        if self.fields:
            kwargs["fields"] = self.fields
        kwargs["type"] = self.type
        kwargs["granularity"] = self.granularity
        return path, self.expressions, kwargs

    def clone(self):
        """Create a copy of this Index."""
        _, args, kwargs = self.deconstruct()
        return self.__class__(*args, **kwargs)

    def set_name_with_model(self, model):
        """
        Generate a unique name for the index.

        The name is divided into 3 parts - table name (12 chars), field name
        (8 chars) and unique hash + suffix (10 chars). Each part is made to
        fit its size by truncating the excess length.
        """
        _, table_name = split_identifier(model._meta.db_table)
        column_names = [
            model._meta.get_field(field_name).column for field_name in self.fields
        ]
        # The length of the parts of the name is based on the default max
        # length of 30 characters.
        hash_data = [table_name] + column_names + [self.suffix]
        self.name = "%s_%s_%s" % (
            table_name[:11],
            column_names[0][:7],
            "%s_%s" % (names_digest(*hash_data, length=6), self.suffix),
        )
        assert len(self.name) <= self.max_name_length, (
            "Index too long for multiple database support. Is self.suffix "
            "longer than 3 characters?"
        )
        if self.name[0] == "_" or self.name[0].isdigit():
            self.name = "D%s" % self.name[1:]

    def __repr__(self):
        return "<%s:%s%s%s%s>" % (
            self.__class__.__name__,
            "" if not self.fields else " fields='%s'" % ", ".join(self.fields),
            ""
            if not self.expressions
            else " expressions='%s'"
            % ", ".join([str(expression) for expression in self.expressions]),
            " type=%s" % self.type,
            " granularity=%s" % self.granularity,
        )

    def __eq__(self, other):
        if self.__class__ == other.__class__:
            return self.deconstruct() == other.deconstruct()
        return NotImplemented


class Type(Func):
    def deconstruct(self):
        path, args, kwargs = super().deconstruct()
        if path.startswith("clickhouse_backend.models.indexes"):
            path = path.replace(
                "clickhouse_backend.models.indexes", "clickhouse_backend.models"
            )
        return path, args, kwargs


class MinMax(Type):
    function = "minmax"
    template = "%(function)s"
    arity = 0


class Set(Type):
    function = "set"
    arity = 1


class NgrambfV1(Type):
    function = "ngrambf_v1"
    arity = 4


class TokenbfV1(Type):
    function = "tokenbf_v1"
    arity = 3


class BloomFilter(Type):
    function = "bloom_filter"



from django.db.models.expressions import Value

from clickhouse_backend.models import fields

from .base import Func

__all__ = [
    "Tuple",
    "tupleElement",
]


class Tuple(Func):
    function = "tuple"

    def _resolve_output_field(self):
        """
        Attempt to infer the output type of the expression.
        """

        sources_iter = (
            source for source in self.get_source_fields() if source is not None
        )
        return fields.TupleField(tuple(sources_iter))


class tupleElement(Func):
    def __init__(self, *expressions, output_field=None, **extra):
        if len(expressions) == 2:
            expressions = (
                expressions[0],
                Value(expressions[1] + 1)
                if isinstance(expressions[1], int)
                else expressions[1],
            )
        elif len(expressions) == 3:
            expressions = (
                expressions[0],
                Value(expressions[1] + 1)
                if isinstance(expressions[1], int)
                else expressions[1],
                expressions[2],
            )
        else:
            raise TypeError(
                "'%s' takes 2 or 3 arguments (%s given)"
                % (
                    self.__class__.__name__,
                    len(expressions),
                )
            )
        super().__init__(*expressions, output_field=output_field, **extra)



import logging
import os
import random
import time

from .base import BaseIDWorker

logger = logging.getLogger(__name__)


class InvalidSystemClock(Exception):
    pass


# 64位ID的划分
WORKER_ID_BITS = 5
DATACENTER_ID_BITS = 5
SEQUENCE_BITS = 12

# 最大取值计算
MAX_WORKER_ID = -1 ^ (-1 << WORKER_ID_BITS)  # 2**5-1 0b11111
MAX_DATACENTER_ID = -1 ^ (-1 << DATACENTER_ID_BITS)

# 移位偏移计算
WORKER_ID_SHIFT = SEQUENCE_BITS
DATACENTER_ID_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS
TIMESTAMP_LEFT_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS + DATACENTER_ID_BITS

# 序号循环掩码
SEQUENCE_MASK = -1 ^ (-1 << SEQUENCE_BITS)

# Twitter元年时间戳
TWEPOCH = 1288834974657


class SnowflakeIDWorker(BaseIDWorker):
    """Snowflake generate universal unique incremental id.
    Copy from https://www.cnblogs.com/oklizz/p/11865750.html
    """

    def __init__(self, datacenter_id, worker_id, sequence=0):
        """
        初始化
        :param datacenter_id: 数据中心（机器区域）ID
        :param worker_id: 机器ID
        :param sequence: 其实序号
        """
        # sanity check
        if worker_id > MAX_WORKER_ID or worker_id < 0:
            raise ValueError("worker_id值越界")

        if datacenter_id > MAX_DATACENTER_ID or datacenter_id < 0:
            raise ValueError("datacenter_id值越界")

        self.worker_id = worker_id
        self.datacenter_id = datacenter_id
        self.sequence = sequence

        self.last_timestamp = -1  # 上次计算的时间戳

    def _gen_timestamp(self):
        """
        生成整数时间戳
        :return:int timestamp
        """
        return int(time.time() * 1000)

    def get_id(self):
        """
        获取新ID
        :return:
        """
        timestamp = self._gen_timestamp()

        # 时钟回拨
        if timestamp < self.last_timestamp:
            logging.error(
                "clock is moving backwards. Rejecting requests until {}".format(
                    self.last_timestamp
                )
            )
            raise InvalidSystemClock

        if timestamp == self.last_timestamp:
            self.sequence = (self.sequence + 1) & SEQUENCE_MASK
            if self.sequence == 0:
                timestamp = self._til_next_millis(self.last_timestamp)
        else:
            self.sequence = 0

        self.last_timestamp = timestamp

        new_id = (
            ((timestamp - TWEPOCH) << TIMESTAMP_LEFT_SHIFT)
            | (self.datacenter_id << DATACENTER_ID_SHIFT)
            | (self.worker_id << WORKER_ID_SHIFT)
            | self.sequence
        )
        return new_id

    def _til_next_millis(self, last_timestamp):
        """
        等到下一毫秒
        """
        timestamp = self._gen_timestamp()
        while timestamp <= last_timestamp:
            timestamp = self._gen_timestamp()
        return timestamp


def get_environ_int(name, minium, maximum, default=None):
    value = os.environ.get(name)
    if value is not None:
        try:
            value = int(value)
        except ValueError:
            raise ValueError(
                '"%s" is not a valid %s. '
                "Hint: set value to an integer between %s and %s"
                % (value, name, minium, maximum)
            )
        if value < minium or value > maximum:
            raise ValueError(
                "%s must be an integer between %s and %s" % (name, minium, maximum)
            )
    else:
        value = default
    return value


def get_default_id_worker():
    """Default id worker for BigAutoField.

    If SnowflakeIDWorker is used, user should ensure worker id
    and datacenter id is unique among all server process.
    Use CLICKHOUSE_WORKER_ID and CLICKHOUSE_DATACENTER_ID
    environment variable to set them.
    :return: BaseIDWorker.
    """

    worker_id = get_environ_int("CLICKHOUSE_WORKER_ID", 0, MAX_WORKER_ID, 0)
    datacenter_id = get_environ_int("CLICKHOUSE_DATACENTER_ID", 0, MAX_DATACENTER_ID)
    if datacenter_id is None:
        datacenter_id = random.randint(0, MAX_DATACENTER_ID)
    return SnowflakeIDWorker(datacenter_id=datacenter_id, worker_id=worker_id)


snowflake_worker = get_default_id_worker()



import ipaddress
from collections.abc import Iterable
from datetime import datetime

from django.core import checks, exceptions
from django.db.models import IntegerChoices, fields
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from clickhouse_backend.validators import MaxBytesValidator

from .array import *  # noqa: F403
from .base import FieldMixin
from .integer import *  # noqa: F403
from .json import *  # noqa: F403
from .map import *  # noqa: F403
from .tuple import *  # noqa: F403

__all__ = [  # noqa: F405
    "Int8Field",
    "UInt8Field",
    "Int16Field",
    "UInt16Field",
    "Int32Field",
    "UInt32Field",
    "Int64Field",
    "UInt64Field",
    "Int128Field",
    "UInt128Field",
    "Int256Field",
    "UInt256Field",
    "Float32Field",
    "Float64Field",
    "DecimalField",
    "BoolField",
    "StringField",
    "FixedStringField",
    "UUIDField",
    "DateField",
    "Date32Field",
    "DateTimeField",
    "DateTime64Field",
    "Enum8Field",
    "Enum16Field",
    "EnumField",
    "IPv4Field",
    "IPv6Field",
    "GenericIPAddressField",
    "ArrayField",
    "TupleField",
    "MapField",
    "JSONField",
]


class Float32Field(FieldMixin, fields.FloatField):
    description = _("Single precision floating point number")

    def __init__(self, *args, low_cardinality=False, **kwargs):
        self.low_cardinality = low_cardinality
        super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return "Float32Field"


class Float64Field(FieldMixin, fields.FloatField):
    description = _("Double precision floating point number")

    def __init__(self, *args, low_cardinality=False, **kwargs):
        self.low_cardinality = low_cardinality
        super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return "Float64Field"


class DecimalField(FieldMixin, fields.DecimalField):
    pass


class BoolField(FieldMixin, fields.BooleanField):
    def __init__(self, *args, low_cardinality=False, **kwargs):
        self.low_cardinality = low_cardinality
        super().__init__(*args, **kwargs)


class StringField(FieldMixin, fields.TextField):
    description = _("Binary string")

    def __init__(self, *args, low_cardinality=False, **kwargs):
        self.low_cardinality = low_cardinality
        super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return "StringField"

    def to_python(self, value):
        if isinstance(value, (str, bytes)) or value is None:
            return value
        return str(value)


class FixedStringField(FieldMixin, fields.TextField):
    description = _("Binary string (up to %(max_bytes)s bytes)")

    def __init__(self, *args, max_bytes=None, low_cardinality=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_bytes = max_bytes
        self.low_cardinality = low_cardinality
        self.validators.append(MaxBytesValidator(self.max_bytes))

    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_max_bytes(),
        ]

    def _check_max_bytes(self):
        if self.max_bytes is None:
            return [
                checks.Error(
                    "FixedStringField must define a 'max_bytes' attribute.",
                    obj=self,
                )
            ]
        elif (
            not isinstance(self.max_bytes, int)
            or isinstance(self.max_bytes, bool)
            or self.max_bytes <= 0
        ):
            return [
                checks.Error(
                    "'max_bytes' must be a positive integer.",
                    obj=self,
                )
            ]
        else:
            return []

    def get_internal_type(self):
        return "FixedStringField"

    def to_python(self, value):
        if isinstance(value, (str, bytes)) or value is None:
            return value
        return str(value)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["max_bytes"] = self.max_bytes
        return name, path, args, kwargs


class UUIDField(FieldMixin, fields.UUIDField):
    def __init__(self, *args, low_cardinality=False, **kwargs):
        self.low_cardinality = low_cardinality
        super().__init__(*args, **kwargs)


# class DateFieldMixin(FieldMixin):
#     def get_db_prep_save(self, value, connection):
#         """Clickhouse Date/Date32 support integer or float value
#         when insert or update (not in where clause)."""
#         if isinstance(value, (float, int)):
#             return value
#         return super().get_db_prep_save(value, connection)


class DateField(FieldMixin, fields.DateField):
    """Support integer or float may cause strange behavior,
    as integer and float are always treated as UTC timestamps,
    clickhouse store them as date in utc, which may not be what you want.
    """

    def __init__(self, *args, low_cardinality=False, **kwargs):
        self.low_cardinality = low_cardinality
        super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return "ClickhouseDateField"


class Date32Field(FieldMixin, fields.DateField):
    def __init__(self, *args, low_cardinality=False, **kwargs):
        self.low_cardinality = low_cardinality
        super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return "Date32Field"


class DateTimeMixin(FieldMixin):
    def get_prep_value(self, value):
        """Clickhouse DateTime('timezone')/DateTime64('timezone')
        support integer or float value in query.

        Integer and float numbers are always treated as UTC timestamps.
        """
        if isinstance(value, (int, float)):
            return value
        return super().get_prep_value(value)


class DateTimeField(DateTimeMixin, fields.DateTimeField):
    def __init__(self, *args, low_cardinality=False, **kwargs):
        self.low_cardinality = low_cardinality
        super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return "ClickhouseDateTimeField"

    def get_prep_value(self, value):
        """For the sake of distinguishing DateTime from DateTime64,
        DateTime value always have microsecond = 0.
        Time string with milliseconds part will cause
        DB::Exception: Cannot convert string 2022-01-01 00:00:00.0 to type DateTime.
        """
        value = super().get_prep_value(value)
        if isinstance(value, datetime):
            value = value.replace(microsecond=0)
        elif isinstance(value, float):
            return int(value)
        return value


class DateTime64Field(DateTimeMixin, fields.DateTimeField):
    DEFAULT_PRECISION = 6

    def __init__(self, *args, precision=DEFAULT_PRECISION, **kwargs):
        super().__init__(*args, **kwargs)
        self.precision = precision

    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_precision(),
        ]

    def _check_precision(self):
        if (
            not isinstance(self.precision, int)
            or isinstance(self.precision, bool)
            or self.precision < 0
            or self.precision > 9
        ):
            return [
                checks.Error(
                    "'precision' must be an integer, valid range: [ 0 : 9 ].",
                    obj=self,
                )
            ]
        else:
            return []

    def get_internal_type(self):
        return "DateTime64Field"

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.precision != self.DEFAULT_PRECISION:
            kwargs["precision"] = self.precision
        return name, path, args, kwargs


class EnumField(FieldMixin, fields.Field):
    description = _("Integer enum value")
    MIN_INT = -32768
    MAX_INT = 32767

    def __init__(self, *args, return_int=True, **kwargs):
        """Use return_int to control whether to get an int or str value when querying from the database."""
        self.return_int = return_int
        super().__init__(*args, **kwargs)

    def _check_choices(self):
        """Note: although clickhouse support arbitrary bytes in Enum name,
        but clickhouse-driver 0.2.5 will raise UnicodeDecodeError when execute query."""

        invalid_errors = [
            checks.Error(
                "'choices' must be an iterable containing (int, str) tuples.",
                obj=self,
                id="fields.E005",
            )
        ]
        if not isinstance(self.choices, Iterable) or isinstance(self.choices, str):
            return invalid_errors

        if not self.choices:
            return [
                checks.Error(
                    f"{self.__class__.__name__} must define a 'choices' attribute.",
                    obj=self,
                )
            ]

        normal_choices = []
        for choice in self.choices:
            try:
                value, name = choice
            except (TypeError, ValueError):
                return invalid_errors
            if not isinstance(value, int) or not isinstance(name, (str, bytes)):
                return invalid_errors
            if value < self.MIN_INT or value > self.MAX_INT:
                return [
                    checks.Error(
                        "'choices' must be in range: [ %s : %s ]."
                        % (self.MIN_INT, self.MAX_INT),
                        obj=self,
                    )
                ]
            if isinstance(name, bytes):
                try:
                    name = name.decode("utf-8")
                except UnicodeDecodeError:
                    pass
            normal_choices.append((value, name))

        normal_choices.sort(key=lambda o: o[0])
        self.choices = normal_choices
        return []

    def get_internal_type(self):
        return "EnumField"

    def db_type(self, connection):
        self._check_backend(connection)
        type_name = connection.data_types[self.get_internal_type()]
        qv = connection.schema_editor().quote_value
        enum_string = ", ".join(
            "%s=%s" % (qv(name), value) for value, name in self.choices
        )
        base_type = "%s(%s)" % (type_name, enum_string)
        return self._nested_type(base_type)

    @cached_property
    def _name_value_map(self):
        return {name: value for value, name in self.choices}

    def value_to_string(self, obj):
        value = self.value_from_object(obj)
        if isinstance(value, int):
            return value
        if isinstance(value, bytes):
            try:
                value = value.decode("utf-8")
            except UnicodeDecodeError:
                pass
        return self._name_value_map[value]

    def get_prep_value(self, value):
        if isinstance(value, bytes):
            try:
                value = value.decode("utf-8")
            except UnicodeDecodeError:
                pass
        elif isinstance(value, IntegerChoices):
            value = value.value
        return value

    def from_db_value(self, value, expression, connection):
        """Clickhouse return enum name for Enum column."""
        if value is None or not self.return_int:
            return value
        return self._name_value_map[value]


class Enum8Field(EnumField):
    description = _("8bit integer enum value")
    MIN_INT = -128
    MAX_INT = 127

    def get_internal_type(self):
        return "Enum8Field"


class Enum16Field(EnumField):
    description = _("16bit integer enum value")

    def get_internal_type(self):
        return "Enum16Field"


class IPv4Field(FieldMixin, fields.GenericIPAddressField):
    description = _("IPv4 address")

    def __init__(self, *args, low_cardinality=False, **kwargs):
        kwargs["protocol"] = "ipv4"
        kwargs["unpack_ipv4"] = False
        self.low_cardinality = low_cardinality
        super().__init__(*args, **kwargs)
        self.max_length = 15

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        for key in ["unpack_ipv4", "protocol", "max_length"]:
            try:
                del kwargs[key]
            except KeyError:
                pass
        return name, path, args, kwargs

    def get_internal_type(self):
        return "IPv4Field"

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return str(value)

    def get_prep_value(self, value):
        value = super(fields.GenericIPAddressField, self).get_prep_value(value)
        if value is None or isinstance(value, ipaddress.IPv4Address):
            return value
        elif isinstance(value, ipaddress.IPv6Address):
            raise exceptions.ValidationError(
                _("This is not a valid IPv4 address."), code="invalid"
            )

        try:
            return ipaddress.IPv4Address(value)
        except ValueError:
            raise exceptions.ValidationError(
                _("This is not a valid IPv4 address."), code="invalid"
            )


class IPv6Field(FieldMixin, fields.GenericIPAddressField):
    description = _("IPv6 address")

    def __init__(self, *args, low_cardinality=False, **kwargs):
        kwargs["protocol"] = "ipv6"
        kwargs["unpack_ipv4"] = False
        self.low_cardinality = low_cardinality
        super().__init__(*args, **kwargs)
        kwargs["max_length"] = 39

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        for key in ["unpack_ipv4", "protocol", "max_length"]:
            try:
                del kwargs[key]
            except KeyError:
                pass
        return name, path, args, kwargs

    def get_internal_type(self):
        return "IPv6Field"

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return str(value)

    def get_prep_value(self, value):
        value = super(fields.GenericIPAddressField, self).get_prep_value(value)
        if value is None:
            return None

        if isinstance(value, str):
            try:
                value = ipaddress.ip_address(value)
            except ValueError:
                raise exceptions.ValidationError(
                    _("This is not a valid IP address."), code="invalid"
                )

        if isinstance(value, ipaddress.IPv4Address):
            value = ipaddress.IPv6Address("::ffff:%s" % value)
        return value


class GenericIPAddressField(FieldMixin, fields.GenericIPAddressField):
    def __init__(self, *args, low_cardinality=False, **kwargs):
        self.low_cardinality = low_cardinality
        super().__init__(*args, **kwargs)

    def db_type(self, connection):
        if self.protocol.lower() == "ipv4":
            return "IPv4"
        else:
            return "IPv6"

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        if self.unpack_ipv4 and value.ipv4_mapped:
            return str(value.ipv4_mapped)
        return str(value)

    def get_prep_value(self, value):
        value = super(fields.GenericIPAddressField, self).get_prep_value(value)
        if value is None:
            return None

        if isinstance(value, str):
            try:
                value = ipaddress.ip_address(value)
            except ValueError:
                raise exceptions.ValidationError(
                    _("This is not a valid IP address."), code="invalid"
                )

        if isinstance(value, ipaddress.IPv4Address) and self.protocol.lower() != "ipv4":
            value = ipaddress.IPv6Address("::ffff:%s" % value)
        elif (
            isinstance(value, ipaddress.IPv6Address) and self.protocol.lower() == "ipv4"
        ):
            if value.ipv4_mapped is None:
                raise exceptions.ValidationError(
                    _("This is not a valid IPv4 address."), code="invalid"
                )
            value = value.ipv4_mapped
        return value



from django.db.models.fields import Field, json

from clickhouse_backend.driver import JSON

from .base import FieldMixin

__all__ = ["JSONField"]


class JSONField(FieldMixin, json.JSONField):
    nullable_allowed = False

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if path.startswith("clickhouse_backend.models.json"):
            path = path.replace(
                "clickhouse_backend.models.json", "clickhouse_backend.models"
            )
        return name, path, args, kwargs

    def from_db_value(self, value, expression, connection):
        return value

    def get_prep_value(self, value):
        # django 4.1 and below dumps value as json string.
        return Field.get_prep_value(self, value)

    def get_db_prep_value(self, value, connection, prepared=False):
        value = super().get_db_prep_value(value, connection, prepared)
        # django 4.1 and below does not call connection.ops.adapt_json_value
        if not isinstance(value, JSON):
            value = JSON(value)
        return value

    def get_db_prep_save(self, value, connection):
        return value



import itertools

from django.core.exceptions import EmptyResultSet
from django.db import NotSupportedError
from django.db.models.fields import AutoFieldMixin
from django.db.models.sql import compiler

from clickhouse_backend import compat
from clickhouse_backend.idworker import id_worker
from clickhouse_backend.models import engines
from clickhouse_backend.models.sql import Query

if compat.dj_ge42:
    from django.core.exceptions import FullResultSet
else:

    class FullResultSet(Exception):
        """A database query predicate is matches everything."""

        pass


# Max rows you can insert using expression as value.
MAX_ROWS_INSERT_USE_EXPRESSION = 1000


class ClickhouseMixin:
    def _add_explain_sql(self, sql, params):
        # Backward compatible for django 3.2
        explain_info = getattr(self.query, "explain_info", None)
        if explain_info:
            prefix, suffix = self.connection.ops.explain_query(
                format=explain_info.format,
                type=explain_info.type,
                **explain_info.options,
            )
            sql = "%s %s" % (prefix, sql.lstrip())
            if suffix:
                sql = "%s %s" % (sql, suffix)
        return sql, params

    def _add_settings_sql(self, sql, params):
        if getattr(self.query, "setting_info", None):
            setting_sql, setting_params = self.connection.ops.settings_sql(
                **self.query.setting_info
            )
            sql = "%s %s" % (sql, setting_sql)
            params = (*params, *setting_params)
        return sql, params

    def _compile_where(self, table):
        try:
            where, params = self.compile(self.query.where)
        except FullResultSet:
            where, params = "", ()
        if where:
            where = where.replace(table + ".", "")
        else:
            where = "TRUE"
        return where, params


class SQLCompiler(ClickhouseMixin, compiler.SQLCompiler):
    def pre_sql_setup(self, with_col_aliases=False):
        """
        Do any necessary class setup immediately prior to producing SQL. This
        is for things that can't necessarily be done in __init__ because we
        might not have all the pieces in place at that time.
        """
        if compat.dj_ge42:
            self.setup_query(with_col_aliases=with_col_aliases)
            (
                self.where,
                self.having,
                self.qualify,
            ) = self.query.where.split_having_qualify(
                must_group_by=self.query.group_by is not None
            )
            if isinstance(self.query, Query):
                (
                    self.prewhere,
                    prehaving,
                    prequalify,
                ) = self.query.prewhere.split_having_qualify(
                    must_group_by=self.query.group_by is not None
                )
            else:
                self.prewhere, prehaving, prequalify = None, None, None
            # Check before ClickHouse complain.
            # DB::Exception: Window function is found in PREWHERE in query. (ILLEGAL_AGGREGATION)
            if prequalify:
                raise NotSupportedError(
                    "Window function is disallowed in the prewhere clause."
                )
        else:
            self.setup_query()
            self.where, self.having = self.query.where.split_having()
            if isinstance(self.query, Query):
                self.prewhere, prehaving = self.query.prewhere.split_having()
            else:
                self.prewhere, prehaving = None, None
        # Check before ClickHouse complain.
        # DB::Exception: Aggregate function is found in PREWHERE in query. (ILLEGAL_AGGREGATION)
        if prehaving:
            raise NotSupportedError(
                "Aggregate function is disallowed in the prewhere clause."
            )
        order_by = self.get_order_by()
        extra_select = self.get_extra_select(order_by, self.select)
        self.has_extra_select = bool(extra_select)
        group_by = self.get_group_by(self.select + extra_select, order_by)
        return extra_select, order_by, group_by

    def as_sql(self, with_limits=True, with_col_aliases=False):
        """
        Create the SQL for this query. Return the SQL string and list of
        parameters.

        If 'with_limits' is False, any limit/offset information is not included
        in the query.
        """
        refcounts_before = self.query.alias_refcount.copy()
        try:
            combinator = self.query.combinator
            if compat.dj_ge42:
                extra_select, order_by, group_by = self.pre_sql_setup(
                    with_col_aliases=with_col_aliases or bool(combinator),
                )
            else:
                extra_select, order_by, group_by = self.pre_sql_setup()
            # Is a LIMIT/OFFSET clause needed?
            with_limit_offset = with_limits and self.query.is_sliced
            if combinator:
                result, params = self.get_combinator_sql(
                    combinator, self.query.combinator_all
                )
            # Django >= 4.2 have this branch
            elif compat.dj_ge42 and self.qualify:
                result, params = self.get_qualify_sql()
                order_by = None
            else:
                distinct_fields, distinct_params = self.get_distinct()
                # This must come after 'select', 'ordering', and 'distinct'
                # (see docstring of get_from_clause() for details).
                from_, f_params = self.get_from_clause()
                try:
                    where, w_params = (
                        self.compile(self.where) if self.where is not None else ("", [])
                    )
                except EmptyResultSet:
                    if compat.dj3 or self.elide_empty:
                        raise
                    # Use a predicate that's always False.
                    where, w_params = "FALSE", []
                except FullResultSet:
                    where, w_params = "", []
                try:
                    having, h_params = (
                        self.compile(self.having)
                        if self.having is not None
                        else ("", [])
                    )
                except FullResultSet:
                    having, h_params = "", []
                # v1.2.0 new feature, support prewhere clause.
                # refer https://clickhouse.com/docs/en/sql-reference/statements/select/prewhere
                try:
                    prewhere, p_params = (
                        self.compile(self.prewhere)
                        if self.prewhere is not None
                        else ("", [])
                    )
                except EmptyResultSet:
                    if compat.dj_ge42 and self.elide_empty:
                        raise
                    # Use a predicate that's always False.
                    prewhere, p_params = "FALSE", []
                except FullResultSet:
                    prewhere, p_params = "", []
                result = ["SELECT"]
                params = []

                if self.query.distinct:
                    distinct_result, distinct_params = self.connection.ops.distinct_sql(
                        distinct_fields,
                        distinct_params,
                    )
                    result += distinct_result
                    params += distinct_params

                out_cols = []
                for _, (s_sql, s_params), alias in self.select + extra_select:
                    if alias:
                        s_sql = "%s AS %s" % (
                            s_sql,
                            self.connection.ops.quote_name(alias),
                        )
                    params.extend(s_params)
                    out_cols.append(s_sql)

                result += [", ".join(out_cols)]
                if from_:
                    result += ["FROM", *from_]
                params.extend(f_params)

                if prewhere:
                    result.append("PREWHERE %s" % prewhere)
                    params.extend(p_params)

                if where:
                    result.append("WHERE %s" % where)
                    params.extend(w_params)

                grouping = []
                for g_sql, g_params in group_by:
                    grouping.append(g_sql)
                    params.extend(g_params)
                if grouping:
                    if distinct_fields:
                        raise NotImplementedError(
                            "annotate() + distinct(fields) is not implemented."
                        )
                    result.append("GROUP BY %s" % ", ".join(grouping))
                    if self._meta_ordering:
                        order_by = None
                if having:
                    result.append("HAVING %s" % having)
                    params.extend(h_params)

            if order_by:
                ordering = []
                for _, (o_sql, o_params, _) in order_by:
                    ordering.append(o_sql)
                    params.extend(o_params)
                order_by_sql = "ORDER BY %s" % ", ".join(ordering)
                result.append(order_by_sql)

            if with_limit_offset:
                result.append(
                    self.connection.ops.limit_offset_sql(
                        self.query.low_mark, self.query.high_mark
                    )
                )

            if self.query.subquery and extra_select:
                # If the query is used as a subquery, the extra selects would
                # result in more columns than the left-hand side expression is
                # expecting. This can happen when a subquery uses a combination
                # of order_by() and distinct(), forcing the ordering expressions
                # to be selected as well. Wrap the query in another subquery
                # to exclude extraneous selects.
                sub_selects = []
                sub_params = []
                for index, (select, _, alias) in enumerate(self.select, start=1):
                    if alias:
                        sub_selects.append(
                            "%s.%s"
                            % (
                                self.connection.ops.quote_name("subquery"),
                                self.connection.ops.quote_name(alias),
                            )
                        )
                    else:
                        select_clone = select.relabeled_clone(
                            {select.alias: "subquery"}
                        )
                        subselect, subparams = select_clone.as_sql(
                            self, self.connection
                        )
                        sub_selects.append(subselect)
                        sub_params.extend(subparams)
                sql = "SELECT %s FROM (%s) subquery" % (
                    ", ".join(sub_selects),
                    " ".join(result),
                )
                params = tuple(sub_params + params)
            else:
                sql = " ".join(result)
                params = tuple(params)
            sql, params = self._add_settings_sql(sql, params)
            sql, params = self._add_explain_sql(sql, params)
            return sql, params
        finally:
            # Finally do cleanup - get rid of the joins we created above.
            self.query.reset_refcounts(refcounts_before)


class SQLInsertCompiler(compiler.SQLInsertCompiler):
    def as_sql(self):
        # We don't need quote_name_unless_alias() here, since these are all
        # going to be column names (so we can avoid the extra overhead).
        qn = self.connection.ops.quote_name
        opts = self.query.get_meta()
        insert_statement = self.connection.ops.insert_statement()

        fields = self.query.fields
        # Generate value for AutoField when needed.
        absent_of_pk = isinstance(opts.pk, AutoFieldMixin) and opts.pk not in fields
        if absent_of_pk:
            fields = fields + [opts.pk]
            for obj in self.query.objs:
                setattr(obj, opts.pk.attname, id_worker.get_id())

        # Check db_default
        if any(compat.field_has_db_default(field) for field in fields):
            from django.db.models.expressions import DatabaseDefault

            other_fields = []
            db_default_fields = []
            for field in fields:
                if compat.field_has_db_default(field):
                    db_default_fields.append(field)
                else:
                    other_fields.append(field)
            fields = other_fields

            # 1. Too many rows should not use mix of const values and db default values, so only check first object.
            if len(self.query.objs) > MAX_ROWS_INSERT_USE_EXPRESSION:
                first_obj = self.query.objs[0]
                for field in db_default_fields:
                    # Field with db_default can be omitted.
                    if isinstance(
                        self.prepare_value(field, self.pre_save_val(field, first_obj)),
                        DatabaseDefault,
                    ):
                        continue
                    fields.append(field)
            else:
                for field in db_default_fields:
                    # Field with db_default can be omitted.
                    if all(
                        isinstance(
                            self.prepare_value(field, self.pre_save_val(field, obj)),
                            DatabaseDefault,
                        )
                        for obj in self.query.objs
                    ):
                        continue
                    fields.append(field)
                # Fields should not be empty.
                if not fields:
                    fields.append(db_default_fields[0])

        result = [
            "%s %s(%s)"
            % (
                insert_statement,
                qn(opts.db_table),
                ", ".join(qn(f.column) for f in fields),
            )
        ]

        value_rows = [
            [
                self.prepare_value(field, self.pre_save_val(field, obj))
                for field in fields
            ]
            for obj in self.query.objs
        ]

        # https://clickhouse.com/docs/en/sql-reference/statements/insert-into
        # If you want to specify SETTINGS for INSERT query then you have to do it before FORMAT clause
        # since everything after FORMAT format_name is treated as data.
        if getattr(self.query, "setting_info", None):
            setting_sql, setting_params = self.connection.ops.settings_sql(
                **self.query.setting_info
            )
            qv = self.connection.schema_editor().quote_value
            result.append((setting_sql % map(qv, setting_params)) % ())

        # If value rows count exceed limitation, raw data is asserted.
        # Refer https://clickhouse-driver.readthedocs.io/en/latest/quickstart.html#inserting-data
        if len(value_rows) > MAX_ROWS_INSERT_USE_EXPRESSION:
            result.append("VALUES")
            params = value_rows
        else:
            placeholder_rows, param_rows = self.assemble_as_sql(fields, value_rows)
            if any(i != "%s" for i in itertools.chain.from_iterable(placeholder_rows)):
                placeholder_rows_sql = (", ".join(row) for row in placeholder_rows)
                values_sql = ", ".join("(%s)" % sql for sql in placeholder_rows_sql)
                result.append("VALUES " + values_sql)
                params = tuple(itertools.chain.from_iterable(param_rows))
            else:
                result.append("VALUES")
                params = param_rows
        return [(" ".join(result), params)]

    def execute_sql(self, returning_fields=None):
        as_sql = self.as_sql()
        with self.connection.cursor() as cursor:
            for sql, params in as_sql:
                cursor.execute(sql, params)
        return []


class SQLDeleteCompiler(ClickhouseMixin, compiler.SQLDeleteCompiler):
    def _as_sql(self, query):
        """
        When execute DELETE and UPDATE query. Clickhouse does not support
        "table"."column" in WHERE clause.
        """
        table = self.quote_name_unless_alias(query.base_table)
        engine = getattr(query.model._meta, "engine", None)
        if isinstance(engine, engines.Distributed):
            cluster = self.quote_name_unless_alias(engine.cluster)
            local_table = self.quote_name_unless_alias(engine.table)
            delete = f"ALTER TABLE {local_table} ON CLUSTER {cluster} DELETE"
        else:
            delete = f"ALTER TABLE {table} DELETE"
        where, params = self._compile_where(table)
        return f"{delete} WHERE {where}", tuple(params)

    def as_sql(self):
        sql, params = super().as_sql()
        sql, params = self._add_settings_sql(sql, params)
        return sql, params


class SQLUpdateCompiler(ClickhouseMixin, compiler.SQLUpdateCompiler):
    def as_sql(self):
        """
        When execute DELETE and UPDATE query. Clickhouse does not support
        "table"."column" in WHERE clause.
        """
        self.pre_sql_setup()
        if not self.query.values:
            return "", ()
        qn = self.quote_name_unless_alias
        values, update_params = [], []
        for field, model, val in self.query.values:
            if hasattr(val, "resolve_expression"):
                val = val.resolve_expression(
                    self.query, allow_joins=False, for_save=True
                )
                if val.contains_aggregate:
                    raise compiler.FieldError(
                        "Aggregate functions are not allowed in this query "
                        "(%s=%r)." % (field.name, val)
                    )
                if val.contains_over_clause:
                    raise compiler.FieldError(
                        "Window expressions are not allowed in this query "
                        "(%s=%r)." % (field.name, val)
                    )
            elif hasattr(val, "prepare_database_save"):
                if field.remote_field:
                    val = field.get_db_prep_value(
                        val.prepare_database_save(field),
                        connection=self.connection,
                    )
                else:
                    raise TypeError(
                        "Tried to update field %s with a model instance, %r. "
                        "Use a value compatible with %s."
                        % (field, val, field.__class__.__name__)
                    )
            else:
                # update params are formatted into query string.
                val = field.get_db_prep_value(val, connection=self.connection)

            # Getting the placeholder for the field.
            if hasattr(field, "get_placeholder"):
                placeholder = field.get_placeholder(val, self, self.connection)
            else:
                placeholder = "%s"
            name = field.column
            if hasattr(val, "as_sql"):
                sql, params = self.compile(val)
                values.append("%s = %s" % (qn(name), placeholder % sql))
                update_params.extend(params)
            elif val is not None:
                values.append("%s = %s" % (qn(name), placeholder))
                update_params.append(val)
            else:
                values.append("%s = NULL" % qn(name))

        # Replace "table"."field" to "field", clickhouse does not support that.
        result = []
        table = qn(self.query.base_table)
        engine = getattr(self.query.model._meta, "engine", None)
        if isinstance(engine, engines.Distributed):
            cluster = qn(engine.cluster)
            local_table = qn(engine.table)
            result.append(f"ALTER TABLE {local_table} ON CLUSTER {cluster} UPDATE")
        else:
            result.append(f"ALTER TABLE {table} UPDATE")
        result.append(", ".join(values).replace(table + ".", ""))
        where, params = self._compile_where(table)
        result.append(f"WHERE {where}")
        params = (*update_params, *params)
        return self._add_settings_sql(" ".join(result), params)


class SQLAggregateCompiler(ClickhouseMixin, compiler.SQLAggregateCompiler):
    def as_sql(self):
        sql, params = super().as_sql()
        sql, params = self._add_settings_sql(sql, params)
        sql, params = self._add_explain_sql(sql, params)
        return sql, params



import re

from django.core.management.commands.inspectdb import Command as DCommand
from django.db import connections

from clickhouse_backend import compat, models
from clickhouse_backend.utils.encoding import ensure_str


class Command(DCommand):
    def handle_inspection(self, options):
        connection = connections[options["database"]]
        if connection.vendor != "clickhouse":
            return super().handle_inspection(options)
        # 'table_name_filter' is a stealth option
        table_name_filter = options.get("table_name_filter")

        def table2model(table_name):
            return re.sub(r"[^a-zA-Z0-9]", "", table_name.title())

        with connection.cursor() as cursor:
            yield "# This is an auto-generated Django model module."
            yield "# You'll have to do the following manually to clean this up:"
            yield "#   * Rearrange models' order"
            yield "#   * Make sure each model has one field with primary_key=True"
            yield (
                "#   * Remove `managed = False` lines if you wish to allow "
                "Django to create, modify, and delete the table"
            )
            yield (
                "# Feel free to rename the models, but don't rename db_table values or "
                "field names."
            )
            yield "from clickhouse_backend import models"
            known_models = []
            # Determine types of tables and/or views to be introspected.
            types = {"t"}
            if options["include_views"]:
                types.add("v")
            table_info = connection.introspection.get_table_list(cursor)
            table_info = {info.name: info for info in table_info if info.type in types}

            for table_name in options["table"] or sorted(name for name in table_info):
                if table_name_filter is not None and callable(table_name_filter):
                    if not table_name_filter(table_name):
                        continue
                try:
                    table_description = connection.introspection.get_table_description(
                        cursor, table_name
                    )
                except Exception as e:
                    yield f"# Unable to inspect table '{table_name}'"
                    yield f"# The error was: {e}"
                    continue

                model_name = table2model(table_name)
                yield ""
                yield ""
                yield f"class {model_name}(models.ClickhouseModel):"
                known_models.append(model_name)
                used_column_names = []  # Holds column names used in the table so far
                column_to_field_name = {}  # Maps column names to names of model fields

                for row in table_description:
                    column_name = row.name

                    (
                        att_name,
                        extra_params,  # Holds Field parameters such as 'db_column'.
                        comment_notes,  # Holds Field notes, to be displayed in a Python comment.
                    ) = self.normalize_col_name(column_name, used_column_names, False)

                    used_column_names.append(att_name)
                    column_to_field_name[column_name] = att_name

                    # Add comment.
                    if (
                        compat.dj_ge42
                        and connection.features.supports_comments
                        and row.comment
                    ):
                        extra_params["db_comment"] = row.comment

                    if extra_params:
                        param = ", ".join(f"{k}={v!r}" for k, v in extra_params.items())
                    else:
                        param = ""

                    field_define = "".join(
                        self.inspect_field_type(row.type_code, param)
                    )
                    field_desc = f"{att_name} = {field_define}"
                    if comment_notes:
                        field_desc += "  # " + " ".join(comment_notes)
                    yield f"    {field_desc}"

                comment = None
                managed_comment = ""
                info = table_info.get(table_name)
                if info:
                    if info.type == "v":
                        managed_comment = "  # Created from a view. Don't remove."
                    if connection.features.supports_comments:
                        comment = info.comment

                yield ""
                yield "    class Meta:"
                yield f"        managed = False{managed_comment}"
                yield f"        db_table = {table_name!r}"
                if compat.dj_ge42 and comment:
                    yield f"        db_table_comment = {comment!r}"

    def inspect_field_type(self, column_type, param=""):
        column_type = ensure_str(column_type)
        # LowCardinality(Int16)
        if column_type.startswith("LowCardinality"):
            param = self.merge_params(param, "low_cardinality=True")
            remain = yield from self.inspect_field_type(column_type[15:], param)
            return remain[1:]
        # Nullable(Int16)
        elif column_type.startswith("Nullable"):
            param = self.merge_params(param, "null=True", "blank=True")
            remain = yield from self.inspect_field_type(column_type[9:], param)
            return remain[1:]
        # FixedString(20)
        elif column_type.startswith("FixedString"):
            i = 12
            while column_type[i].isdigit():
                i += 1
            param = self.merge_params(param, f"max_bytes={column_type[12:i]}")
            yield f"models.FixedStringField({param})"
            return column_type[i + 1 :]
        # DateTime64(6, 'UTC') or DateTime64(9)
        elif column_type.startswith("DateTime64"):
            if int(column_type[11]) != models.DateTime64Field.DEFAULT_PRECISION:
                param = self.merge_params(param, f"precision={column_type[11]}")
            yield f"models.DateTime64Field({param})"

            if column_type[12] == ",":
                i = 15
                while column_type[i] != "'":
                    i += 1
                return column_type[i + 2 :]
            return column_type[13:]
        # DateTime('UTC') or DateTime
        elif column_type.startswith("DateTime"):
            yield f"models.DateTimeField({param})"
            if len(column_type) > 8 and column_type[8] == "(":
                i = 10
                while column_type[i] != "'":
                    i += 1
                    return column_type[i + 2 :]
            return column_type[8:]
        # Decimal(9, 3)
        elif column_type.startswith("Decimal"):
            i = 8
            while column_type[i].isdigit():
                i += 1
            max_digits = f"max_digits={column_type[8:i]}"
            i += 2
            j = i
            while column_type[i].isdigit():
                i += 1
            decimal_places = f"decimal_places={column_type[j:i]}"
            param = self.merge_params(param, max_digits, decimal_places)
            yield f"models.DecimalField({param})"
            return column_type[i + 1 :]
        # Enum8('a' = 1, 'b' = 2)
        elif column_type.startswith("Enum"):
            i = 4
            while column_type[i].isdigit():
                i += 1
            typ = column_type[:i]
            choices = []
            name, value, remain = self.consume_enum_choice(column_type[i + 1 :])
            choices.append(f"({value}, {name})")
            while remain[0] != ")":
                name, value, remain = self.consume_enum_choice(remain[2:])
                choices.append(f"({value}, {name})")
            param = self.merge_params(param, f"choices=[{', '.join(choices)}]")
            yield f"models.{typ}Field({param})"
            return remain[1:]
        # Array(Tuple(String, Enum8('a' = 1, 'b' = 2)))
        elif column_type.startswith("Array"):
            yield "models.ArrayField("
            remain = yield from self.inspect_field_type(column_type[6:])
            if param:
                yield f", {param}"
            yield ")"
            return remain[1:]
        # Tuple(String, Enum8('a' = 1, 'b' = 2))
        elif column_type.startswith("Tuple"):
            yield "models.TupleField(["
            remain = yield from self.inspect_field_type(column_type[6:])
            while remain[0] == ",":
                yield ", "
                remain = yield from self.inspect_field_type(remain[2:])
            yield "]"
            if param:
                yield f", {param}"
            yield ")"
            return remain[1:]
        # Map(String, Int8)
        elif column_type.startswith("Map"):
            yield "models.MapField("
            remain = yield from self.inspect_field_type(column_type[4:])
            yield ", "
            remain = yield from self.inspect_field_type(remain[2:])
            if param:
                yield f", {param}"
            yield ")"
            return remain[1:]
        elif column_type.startswith("Object('json')"):
            yield f"models.JSONField({param})"
            return column_type[14:]

        i = 0
        length = len(column_type)
        while i < length and column_type[i].isalnum():
            i += 1
        yield f"models.{column_type[:i]}Field({param})"
        return column_type[i:]

    def consume_enum_choice(self, s):
        # 'a' = 1
        has_bytes = False
        i = 1
        while True:
            if s[i] == "\\":  # escape char
                if s[i + 1] == "x":
                    has_bytes = True
                i += 2
                continue
            if s[i] == "'":
                break
            i += 1
        i += 1
        name = s[:i]
        # try decoding bytes to utf8 string.
        if has_bytes:
            try:
                decoded = eval(f"b{name}.decode('utf-8')")
            except UnicodeDecodeError:
                name = f"b{name}"
            else:
                name = repr(decoded)

        i += 3
        j = i
        while s[i].isdigit():
            i += 1
        value = s[j:i]
        return name, value, s[i:]

    def merge_params(self, *params):
        return ", ".join(filter(None, params))



from django.db import models
from django.db.migrations import state
from django.db.models import options

from .query import QuerySet
from .sql import Query
from clickhouse_backend import compat

options.DEFAULT_NAMES = (
    *options.DEFAULT_NAMES,
    # Clickhouse features
    "engine",
    "cluster",
)
# Also monkey patch state.DEFAULT_NAMES, this makes new option names contained in migrations.
state.DEFAULT_NAMES = options.DEFAULT_NAMES


class ClickhouseManager(models.Manager):
    _queryset_class = QuerySet

    def get_queryset(self):
        """
        User defined Query and QuerySet class that support clickhouse particular query.
        """
        return self._queryset_class(
            model=self.model, query=Query(self.model), using=self._db, hints=self._hints
        )

    def settings(self, **kwargs):
        return self.get_queryset().settings(**kwargs)

    def prewhere(self, *args, **kwargs):
        return self.get_queryset().prewhere(*args, **kwargs)

    def datetimes(self, *args, **kwargs):
        return self.get_queryset().datetimes(*args, **kwargs)


class ClickhouseModel(models.Model):
    objects = ClickhouseManager()
    _overwrite_base_manager = ClickhouseManager()

    class Meta:
        abstract = True
        base_manager_name = "_overwrite_base_manager"

    if compat.dj_ge6:

        def _do_update(
            self,
            base_qs,
            using,
            pk_val,
            values,
            update_fields,
            forced_update,
            returning_fields,
        ):
            """
            Try to update the model. Return True if the model was updated (if an
            update query was done and a matching row was found in the DB).
            """
            filtered = base_qs.filter(pk=pk_val)
            if not values:
                if update_fields is not None or filtered.exists():
                    return [()]
                return []
            return filtered._update(values, returning_fields)

    else:

        def _do_update(
            self, base_qs, using, pk_val, values, update_fields, forced_update
        ):
            """
            Try to update the model. Return True if the model was updated (if an
            update query was done and a matching row was found in the DB).
            """
            filtered = base_qs.filter(pk=pk_val)
            if not values:
                # We can end up here when saving a model in inheritance chain where
                # update_fields doesn't target any field in current model. In that
                # case we just say the update succeeded. Another case ending up here
                # is a model with just PK - in that case check that the PK still
                # exists.
                return update_fields is not None or filtered.exists()
            return filtered._update(values) > 0



from datetime import date, datetime, time, timezone
from enum import Enum
from ipaddress import IPv4Address, IPv6Address
from itertools import chain
from typing import Dict, Sequence, Union
from uuid import UUID

from clickhouse_driver.util import escape

from . import types

Params = Union[Sequence, Dict]


def escape_datetime(item: datetime, context):
    """Clickhouse backend always treats DateTime[64] with timezone as in UTC timezone.

    DateTime value does not support microsecond part,
    clickhouse_backend.models.DateTimeField will set microsecond to zero.
    As integer and float are always treated as UTC timestamps,
    it is required to convert a naive datetime to an utc timestamp.
    """
    if item.tzinfo is None:
        return item.timestamp()

    item = item.astimezone(timezone.utc)
    if item.microsecond == 0:
        return "'%s'" % item.strftime("%Y-%m-%d %H:%M:%S")
    else:
        return "'%s'" % item.strftime("%Y-%m-%d %H:%M:%S.%f")


def escape_binary(item: bytes, context):
    # b"\x00F '\xfe" ->   '\x00F \'\xfe'
    b2s = str(item)
    if b2s[1] == '"':
        return "'%s'" % b2s[2:-1].replace("'", "\\'")
    return b2s[1:]


@escape.maybe_enquote_for_server
def escape_param(item, context, for_server=False):
    if item is None:
        return "NULL"

    elif isinstance(item, datetime):
        return escape_datetime(item, context)

    elif isinstance(item, date):
        return "'%s'" % item.strftime("%Y-%m-%d")

    elif isinstance(item, time):
        return "'%s'" % item.strftime("%H:%M:%S")

    elif isinstance(item, str):
        # We need double escaping for server-side parameters.
        if for_server:
            item = "".join(escape.escape_chars_map.get(c, c) for c in item)
        return "'%s'" % "".join(escape.escape_chars_map.get(c, c) for c in item)

    elif isinstance(item, list):
        return "[%s]" % ",".join(
            str(escape_param(x, context, for_server=for_server)) for x in item
        )

    elif isinstance(item, tuple):
        return "tuple(%s)" % ",".join(
            str(escape_param(x, context, for_server=for_server)) for x in item
        )

    elif isinstance(item, dict):
        return "map(%s)" % ",".join(
            str(escape_param(x, context, for_server=for_server))
            for x in chain.from_iterable(item.items())
        )

    elif isinstance(item, Enum):
        return escape_param(item.value, context, for_server=for_server)

    elif isinstance(item, (UUID, IPv4Address, IPv6Address)):
        return "'%s'" % str(item)

    elif isinstance(item, types.Binary):
        return escape_binary(item, context)

    elif isinstance(item, types.JSON):
        value = item.value
        if isinstance(value, list):
            return escape_param(
                [types.JSON(v) for v in value], context, for_server=for_server
            )
        elif isinstance(value, dict):
            return escape_param(
                tuple(types.JSON(v) for v in value.values()),
                context,
                for_server=for_server,
            )
        else:
            return escape_param(value, context, for_server=for_server)

    else:
        return item


def escape_params(params: Params, context: Dict, for_server=False) -> Params:
    """Escape param to qualified string representation.

    This function is not used in INSERT INTO queries.
    """
    if isinstance(params, dict):
        escaped = {
            key: escape_param(value, context, for_server=for_server)
            for key, value in params.items()
        }
    else:
        escaped = tuple(
            escape_param(value, context, for_server=for_server) for value in params
        )

    return escaped



import datetime
import functools
import os
import subprocess


# Borrowed from django.utils.version
def get_version(version):
    """Return a PEP 440-compliant version number from VERSION."""

    # Now build the two parts of the version number:
    # main = X.Y[.Z]
    # sub = .devN - for pre-alpha releases
    #     | {a|b|rc}N - for alpha, beta, and rc releases

    main = get_main_version(version)

    sub = ""
    if version[3] == "alpha" and version[4] == 0:
        git_changeset = get_git_changeset()
        if git_changeset:
            sub = ".dev%s" % git_changeset

    elif version[3] != "final":
        mapping = {"alpha": "a", "beta": "b", "rc": "rc"}
        sub = mapping[version[3]] + str(version[4])

    return main + sub


def get_main_version(version):
    """Return main version (X.Y[.Z]) from VERSION."""
    parts = 2 if version[2] == 0 else 3
    return ".".join(str(x) for x in version[:parts])


@functools.lru_cache(maxsize=128)
def get_git_changeset():
    """Return a numeric identifier of the latest git changeset.

    The result is the UTC timestamp of the changeset in YYYYMMDDHHMMSS format.
    This value isn't guaranteed to be unique, but collisions are very unlikely,
    so it's sufficient for generating the development version numbers.
    """
    # Repository may not be found if __file__ is undefined, e.g. in a frozen
    # module.
    if "__file__" not in globals():
        return None
    repo_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    git_log = subprocess.run(
        "git log --pretty=format:%ct --quiet -1 HEAD",
        capture_output=True,
        shell=True,
        cwd=repo_dir,
        text=True,
    )
    timestamp = git_log.stdout
    tz = datetime.timezone.utc
    try:
        timestamp = datetime.datetime.fromtimestamp(int(timestamp), tz=tz)
    except ValueError:
        return None
    return timestamp.strftime("%Y%m%d%H%M%S")



from django.apps.registry import Apps
from django.db import DatabaseError
from django.db import models as django_models
from django.db.migrations import Migration
from django.db.migrations.exceptions import IrreversibleError, MigrationSchemaMissing
from django.db.migrations.operations.fields import FieldOperation
from django.db.migrations.operations.models import (
    DeleteModel,
    IndexOperation,
    ModelOperation,
    RenameModel,
)
from django.db.migrations.recorder import MigrationRecorder
from django.utils.timezone import now

__all__ = ["patch_migrations", "patch_migration_recorder", "patch_migration"]


def _should_distribute_migrations(connection):
    """
    Check if the connection is configured for distributed migrations.
    """
    return getattr(connection, "distributed_migrations", False) and getattr(
        connection, "migration_cluster", None
    )


def _get_model_table_name(connection):
    """
    Return the name of the table that will be used by the MigrationRecorder.
    If distributed migrations are enabled, return the distributed table name.
    Otherwise, return the regular django_migrations table name.
    """
    if _should_distribute_migrations(connection):
        return "distributed_django_migrations"
    return "django_migrations"


def _check_replicas(connection):
    """
    Check if the connection has replicas configured for the migration cluster.
    """
    if hasattr(connection, "has_replicas"):
        return connection.has_replicas

    with connection.cursor() as cursor:
        replica_count = _get_replicas(connection.migration_cluster, cursor)
    return replica_count >= 1


def _get_replicas(cluster_name, cursor):
    cursor.execute(
        "select replica_num from system.clusters where cluster=%s", [cluster_name]
    )
    res = cursor.fetchone()
    if not res:
        return 0
    return res[0]


def patch_migrations():
    patch_migration_recorder()
    patch_migration()


def patch_migration_recorder():
    def Migration(self):
        """
        Lazy load to avoid AppRegistryNotReady if installed apps import
        MigrationRecorder.
        """
        if self._migration_class is None:
            if self.connection.vendor == "clickhouse":
                from clickhouse_backend import models
                from clickhouse_backend.models import currentDatabase

                # Only create a distributed migration model if the connection
                # has distributed migrations enabled and a migration cluster is set.
                # otherwise, create a regular merge tree.
                if _should_distribute_migrations(self.connection):
                    has_replicas = _check_replicas(self.connection)

                    Engine = models.MergeTree
                    if has_replicas:
                        Engine = models.ReplicatedMergeTree

                    self.connection.has_replicas = has_replicas

                    class _Migration(models.ClickhouseModel):
                        app = models.StringField(max_length=255)
                        name = models.StringField(max_length=255)
                        applied = models.DateTime64Field(default=now)
                        deleted = models.BoolField(default=False)

                        class Meta:
                            apps = Apps()
                            app_label = "migrations"
                            db_table = "django_migrations"
                            engine = Engine(order_by=("app", "name"))
                            cluster = self.connection.migration_cluster

                        def __str__(self):
                            return "Migration %s for %s" % (self.name, self.app)

                    class Migration(models.ClickhouseModel):
                        app = models.StringField(max_length=255)
                        name = models.StringField(max_length=255)
                        applied = models.DateTime64Field(default=now)
                        deleted = models.BoolField(default=False)

                        class Meta:
                            apps = Apps()
                            app_label = "migrations"
                            db_table = _get_model_table_name(self.connection)
                            engine = models.Distributed(
                                self.connection.migration_cluster,
                                currentDatabase(),
                                _Migration._meta.db_table,
                                models.Rand(),
                            )
                            cluster = self.connection.migration_cluster

                    Migration._meta.local_model_class = _Migration

                else:

                    class Migration(models.ClickhouseModel):
                        app = models.StringField(max_length=255)
                        name = models.StringField(max_length=255)
                        applied = models.DateTime64Field(default=now)
                        deleted = models.BoolField(default=False)

                        class Meta:
                            apps = Apps()
                            app_label = "migrations"
                            db_table = _get_model_table_name(self.connection)
                            engine = models.MergeTree(order_by=("app", "name"))
                            cluster = getattr(
                                self.connection, "migration_cluster", None
                            )

                        def __str__(self):
                            return "Migration %s for %s" % (self.name, self.app)

            else:

                class Migration(django_models.Model):
                    app = django_models.CharField(max_length=255)
                    name = django_models.CharField(max_length=255)
                    applied = django_models.DateTimeField(default=now)

                    class Meta:
                        apps = Apps()
                        app_label = "migrations"
                        db_table = "django_migrations"

                    def __str__(self):
                        return "Migration %s for %s" % (self.name, self.app)

            self._migration_class = Migration
        return self._migration_class

    def has_table(self):
        """Return True if the django_migrations table exists."""
        # Assert migration table won't be deleted once created.
        if not getattr(self, "_has_table", False):
            with self.connection.cursor() as cursor:
                table = self.Migration._meta.db_table
                tables = self.connection.introspection.table_names(cursor)
                self._has_table = table in tables
                if self._has_table and self.connection.vendor == "clickhouse":
                    # fix https://github.com/jayvynl/django-clickhouse-backend/issues/51
                    cursor.execute(
                        f"ALTER table {table} ADD COLUMN IF NOT EXISTS deleted Bool"
                    )
        return self._has_table

    def ensure_schema(self):
        """Ensure the table exists and has the correct schema."""
        # If the table's there, that's fine - we've never changed its schema
        # in the codebase.
        if self.has_table():
            return

        # In case of distributed migrations, we need to ensure the local model exists first and
        # then create the distributed model.
        try:
            with self.connection.schema_editor() as editor:
                if (
                    editor.connection.vendor == "clickhouse"
                    and _should_distribute_migrations(editor.connection)
                ):
                    with editor.connection.cursor() as cursor:
                        tables = editor.connection.introspection.table_names(cursor)
                    local_model_class = self.Migration._meta.local_model_class
                    local_table = local_model_class._meta.db_table
                    if local_table not in tables:
                        # Create the local model first
                        editor.create_model(self.Migration._meta.local_model_class)

                editor.create_model(self.Migration)
        except DatabaseError as exc:
            raise MigrationSchemaMissing(
                "Unable to create the django_migrations table (%s)" % exc
            )

    def migration_qs(self):
        if self.connection.vendor == "clickhouse":
            return self.Migration.objects.using(self.connection.alias).filter(
                deleted=False
            )
        return self.Migration.objects.using(self.connection.alias)

    def record_applied(self, app, name):
        """Record that a migration was applied."""
        self.ensure_schema()
        if self.connection.vendor == "clickhouse" and (
            self.Migration.objects.using(self.connection.alias)
            .filter(app=app, name=name)
            .exists()
        ):
            self.Migration.objects.using(self.connection.alias).filter(
                app=app, name=name
            ).settings(mutations_sync=1).update(deleted=False)
        else:
            self.migration_qs.create(app=app, name=name)

    def record_unapplied(self, app, name):
        """Record that a migration was unapplied."""
        self.ensure_schema()
        if self.connection.vendor == "clickhouse":
            self.migration_qs.filter(app=app, name=name).settings(
                mutations_sync=1
            ).update(deleted=True)
        else:
            self.migration_qs.filter(app=app, name=name).delete()

    def flush(self):
        """Delete all migration records. Useful for testing migrations."""
        if self.connection.vendor == "clickhouse":
            self.migration_qs.settings(mutations_sync=1).delete()
        else:
            self.migration_qs.all().delete()

    MigrationRecorder.Migration = property(Migration)
    MigrationRecorder.has_table = has_table
    MigrationRecorder.ensure_schema = ensure_schema
    MigrationRecorder.migration_qs = property(migration_qs)
    MigrationRecorder.record_applied = record_applied
    MigrationRecorder.record_unapplied = record_unapplied
    MigrationRecorder.flush = flush


def patch_migration():
    def apply(self, project_state, schema_editor, collect_sql=False):
        """
        Take a project_state representing all migrations prior to this one
        and a schema_editor for a live database and apply the migration
        in a forwards order.

        Return the resulting project state for efficient reuse by following
        Migrations.
        """
        applied_on_remote = False
        if getattr(schema_editor.connection, "migration_cluster", None):
            _table = _get_model_table_name(schema_editor.connection)

            with schema_editor.connection.cursor() as cursor:
                cursor.execute(
                    "select EXISTS(select 1 from clusterAllReplicas(%s, currentDatabase(), %s)"
                    " where app=%s and name=%s and deleted=false)",
                    [
                        schema_editor.connection.migration_cluster,
                        _table,
                        self.app_label,
                        self.name,
                    ],
                )
                (applied_on_remote,) = cursor.fetchone()
        for operation in self.operations:
            # If this operation cannot be represented as SQL, place a comment
            # there instead
            if collect_sql:
                schema_editor.collected_sql.append("--")
                schema_editor.collected_sql.append("-- %s" % operation.describe())
                schema_editor.collected_sql.append("--")
                if not operation.reduces_to_sql:
                    schema_editor.collected_sql.append(
                        "-- THIS OPERATION CANNOT BE WRITTEN AS SQL"
                    )
                    continue
                collected_sql_before = len(schema_editor.collected_sql)
            # Save the state before the operation has run
            old_state = project_state.clone()
            operation.state_forwards(self.app_label, project_state)

            # Run the operation
            # Ensure queries on cluster are only executed once.
            model_name = None
            skip_database_forwards = False
            if isinstance(operation, (IndexOperation, FieldOperation)):
                model_name = operation.model_name_lower
            elif isinstance(operation, ModelOperation):
                model_name = operation.name_lower
            if model_name:
                if isinstance(operation, (RenameModel, DeleteModel)):
                    model_state = old_state.models[self.app_label, model_name]
                else:
                    model_state = project_state.models[self.app_label, model_name]
                if model_state.options.get("cluster") and applied_on_remote:
                    skip_database_forwards = True
            if not skip_database_forwards:
                operation.database_forwards(
                    self.app_label, schema_editor, old_state, project_state
                )
            if collect_sql and collected_sql_before == len(schema_editor.collected_sql):
                schema_editor.collected_sql.append("-- (no-op)")
        return project_state

    def unapply(self, project_state, schema_editor, collect_sql=False):
        """
        Take a project_state representing all migrations prior to this one
        and a schema_editor for a live database and apply the migration
        in a reverse order.

        The backwards migration process consists of two phases:

        1. The intermediate states from right before the first until right
           after the last operation inside this migration are preserved.
        2. The operations are applied in reverse order using the states
           recorded in step 1.
        """
        unapplied_on_remote = False
        if getattr(schema_editor.connection, "migration_cluster", None):
            _table = _get_model_table_name(schema_editor.connection)

            with schema_editor.connection.cursor() as cursor:
                cursor.execute(
                    "select EXISTS(select 1 from clusterAllReplicas(%s, currentDatabase(), %s)"
                    " where app=%s and name=%s and deleted=true)",
                    [
                        schema_editor.connection.migration_cluster,
                        _table,
                        self.app_label,
                        self.name,
                    ],
                )
                (unapplied_on_remote,) = cursor.fetchone()
        # Construct all the intermediate states we need for a reverse migration
        to_run = []
        new_state = project_state
        # Phase 1
        for operation in self.operations:
            # If it's irreversible, error out
            if not operation.reversible:
                raise IrreversibleError(
                    "Operation %s in %s is not reversible" % (operation, self)
                )
            # Preserve new state from previous run to not tamper the same state
            # over all operations
            new_state = new_state.clone()
            old_state = new_state.clone()
            operation.state_forwards(self.app_label, new_state)
            to_run.insert(0, (operation, old_state, new_state))

        # Phase 2
        for operation, to_state, from_state in to_run:
            if collect_sql:
                schema_editor.collected_sql.append("--")
                schema_editor.collected_sql.append("-- %s" % operation.describe())
                schema_editor.collected_sql.append("--")
                if not operation.reduces_to_sql:
                    schema_editor.collected_sql.append(
                        "-- THIS OPERATION CANNOT BE WRITTEN AS SQL"
                    )
                    continue
                collected_sql_before = len(schema_editor.collected_sql)
            # Ensure queries on cluster are only executed once.
            model_name = None
            skip_database_backwards = False
            if isinstance(operation, (IndexOperation, FieldOperation)):
                model_name = operation.model_name_lower
            elif isinstance(operation, ModelOperation):
                model_name = operation.name_lower
            if model_name:
                if isinstance(operation, (RenameModel, DeleteModel)):
                    model_state = to_state.models[self.app_label, model_name]
                else:
                    model_state = from_state.models[self.app_label, model_name]
                if model_state.options.get("cluster") and unapplied_on_remote:
                    skip_database_backwards = True
            if not skip_database_backwards:
                operation.database_backwards(
                    self.app_label, schema_editor, from_state, to_state
                )
            if collect_sql and collected_sql_before == len(schema_editor.collected_sql):
                schema_editor.collected_sql.append("-- (no-op)")
        return project_state

    Migration.apply = apply
    Migration.unapply = unapply



from django.db.models import Value

from clickhouse_backend.models import fields

from .base import Func

__all__ = [
    "currentDatabase",
    "hostName",
    "generateSerialID",
]


class currentDatabase(Func):
    arity = 0
    output_field = fields.StringField()


class hostName(Func):
    arity = 0
    output_field = fields.StringField()


class generateSerialID(Func):
    """
    https://clickhouse.com/docs/sql-reference/functions/other-functions#generateSerialID
    """

    output_field = fields.UInt64Field()

    def __init__(self, *expressions):
        arity = len(expressions)
        if arity < 1 or arity > 2:
            raise TypeError(
                "'%s' takes 1 or 2 arguments (%s given)"
                % (
                    self.__class__.__name__,
                    len(expressions),
                )
            )

        if isinstance(expressions[0], str):
            expressions = (Value(expressions[0]), *expressions[1:])
        super().__init__(*expressions)



from django.db.models.fields import json


def key_transform_as_clickhouse(self, compiler, connection):
    lhs, params, key_transforms = self.preprocess_lhs(compiler, connection)
    sql = lhs
    params = list(params)
    for key in key_transforms:
        if key.isdigit():
            sql = f"{sql}[%s]"
            params.append(int(key) + 1)
        else:
            sql = f"tupleElement({sql}, %s)"
            params.append(key)
    return sql, tuple(params)


def patch_jsonfield():
    json.KeyTransform.as_clickhouse = key_transform_as_clickhouse



from django.conf import settings
from django.utils.functional import SimpleLazyObject
from django.utils.module_loading import import_string

from .base import BaseIDWorker
from .snowflake import SnowflakeIDWorker

__all__ = ["BaseIDWorker", "SnowflakeIDWorker", "id_worker"]

id_worker = SimpleLazyObject(
    lambda: import_string(
        getattr(
            settings,
            "CLICKHOUSE_ID_WORKER",
            "clickhouse_backend.idworker.snowflake.snowflake_worker",
        )
    )
)



from clickhouse_driver import defines
from clickhouse_driver.dbapi import __all__, apilevel, paramstyle, threadsafety  # NOQA
from clickhouse_driver.dbapi.errors import (  # NOQA
    DatabaseError,
    DataError,
    Error,
    IntegrityError,
    InterfaceError,
    InternalError,
    NotSupportedError,
    OperationalError,
    ProgrammingError,
    Warning,
)

from .connection import Connection

# Binary is compatible for django's BinaryField.
from .types import JSON, Binary  # NOQA


def connect(
    dsn=None,
    host=None,
    user=defines.DEFAULT_USER,
    password=defines.DEFAULT_PASSWORD,
    port=defines.DEFAULT_PORT,
    database=defines.DEFAULT_DATABASE,
    **kwargs,
):
    """
    Support dict type params in INSERT query and support connection pool.
    """

    if dsn is None and host is None:
        raise ValueError("host or dsn is required")

    return Connection(
        dsn=dsn,
        user=user,
        password=password,
        host=host,
        port=port,
        database=database,
        **kwargs,
    )



import re
from collections import namedtuple

from django.db.backends.base.introspection import BaseDatabaseIntrospection
from django.db.backends.base.introspection import FieldInfo as BaseFieldInfo
from django.db.backends.base.introspection import TableInfo as BaseTableInfo
from django.utils.functional import cached_property

from clickhouse_backend.utils.encoding import ensure_str

FieldInfo = namedtuple("FieldInfo", BaseFieldInfo._fields + ("comment",))
TableInfo = namedtuple("TableInfo", BaseTableInfo._fields + ("comment",))

constraint_pattern = re.compile(
    r"CONSTRAINT (`)?((?(1)(?:\\.|[^`])+|\S+))(?(1)`|) (CHECK .+?),?\n"
)
index_pattern = re.compile(
    r"INDEX (`)?((?(1)(?:\\.|[^`])+|\S+))(?(1)`|) (.+? TYPE ([a-zA-Z_][0-9a-zA-Z_]*)\(.+?\) GRANULARITY \d+)"
)


class DatabaseIntrospection(BaseDatabaseIntrospection):
    def get_field_type(self, data_type, description):
        if data_type.startswith("LowCardinality"):  # LowCardinality(Int16)
            data_type = data_type[15:-1]
        if data_type.startswith("Nullable"):  # Nullable(Int16)
            data_type = data_type[9:-1]
        if data_type.startswith("FixedString"):  # FixedString(20)
            return "FixedStringField"
        elif data_type.startswith("DateTime64"):  # DateTime64(6, 'UTC')
            return "DateTime64Field"
        elif data_type.startswith("Decimal"):  # Decimal(9, 3)
            return "DecimalField"
        elif data_type.startswith("Enum8"):
            return "Enum8Field"
        elif data_type.startswith("Enum16"):
            return "Enum16Field"
        elif data_type.startswith("Enum"):
            return "EnumField"
        elif data_type.startswith("Array"):
            return "ArrayField"
        elif data_type.startswith("Tuple"):
            return "TupleField"
        elif data_type.startswith("Map"):
            return "MapField"
        elif data_type == "Object('json')":
            return "JSONField"

        return f"{data_type}Field"  # Int8

    def get_table_list(self, cursor):
        """Return a list of table and view names in the current database."""
        cursor.execute(
            """
            SELECT name,
            if(engine LIKE '%%View', 'v', 't'),
            comment
            FROM system.tables
            WHERE database = currentDatabase()
            AND NOT is_temporary
            AND engine NOT LIKE 'System%%'
            AND has_own_data
        """
        )
        return [TableInfo(*row) for row in cursor.fetchall()]

    def get_table_description(self, cursor, table_name):
        """Return a description of the table."""
        cursor.execute(
            """
            SELECT name, type, character_octet_length, character_octet_length,
            coalesce(numeric_precision, datetime_precision),
            numeric_scale, type LIKE 'Nullable(%%)', default_expression, NULL, comment
            FROM system.columns
            WHERE database = currentDatabase() AND table = %s
        """,
            [table_name],
        )
        return [FieldInfo(*line) for line in cursor.fetchall()]

    def get_constraints(self, cursor, table_name):
        """Retrieve any constraints and indexes."""
        constraints = {}
        # No way to get structured data, parse from SHOW CREATE TABLE.
        # https://clickhouse.com/docs/en/sql-reference/statements/show#show-create-table
        cursor.execute('SHOW CREATE TABLE "%s"' % table_name)
        (table_sql,) = cursor.fetchone()
        table_sql = ensure_str(table_sql)
        for backtick, name, definition in constraint_pattern.findall(table_sql):
            constraints[name] = {
                "columns": [],
                "primary_key": False,
                "unique": False,
                "foreign_key": None,
                "check": True,
                "index": False,
                "definition": definition,
                "options": None,
            }

        for backtick, name, definition, type_ in index_pattern.findall(table_sql):
            constraints[name] = {
                "columns": [],
                "orders": [],
                "primary_key": False,
                "unique": False,
                "foreign_key": None,
                "check": False,
                "index": True,
                "type": type_,
                "definition": definition,
                "options": None,
            }
        return constraints

    @cached_property
    def settings(self) -> set:
        """
        Get all available settings.
        """
        with self.connection.cursor() as cursor:
            cursor.execute("SELECT name from system.settings")
            rows = cursor.fetchall()
        return {row[0] for row in rows}



from django.db.models.expressions import Value

from clickhouse_backend.models import fields

from .base import Func
from .tuples import Tuple

__all__ = [
    "halfMD5",
    "MD4",
    "MD5",
    "sipHash64",
    "sipHash64Keyed",
    "sipHash128",
    "sipHash128Keyed",
    "sipHash128Reference",
    "sipHash128ReferenceKeyed",
    "cityHash64",
    "intHash32",
    "intHash64",
    "SHA1",
    "SHA224",
    "SHA256",
    "SHA512",
    "SHA512_256",
    "BLAKE3",
    "URLHash",
    "farmFingerprint64",
    "farmHash64",
]


class halfMD5(Func):
    output_field = fields.UInt64Field()


class MD4(Func):
    arity = 1
    output_field = fields.FixedStringField(max_bytes=16)


class MD5(Func):
    arity = 1
    output_field = fields.FixedStringField(max_bytes=16)


class sipHash64(Func):
    output_field = fields.UInt64Field()


class sipHash64Keyed(Func):
    template = "%(function)s((%(k0)s, %(k1)s), %(expressions)s)"
    output_field = fields.UInt64Field()

    def __init__(self, k0, k1, *expressions):
        super().__init__(Tuple(k0, k1), *expressions)


class sipHash128(Func):
    output_field = fields.FixedStringField(max_bytes=16)


class sipHash128Keyed(sipHash64Keyed):
    output_field = fields.FixedStringField(max_bytes=16)


class sipHash128Reference(Func):
    output_field = fields.FixedStringField(max_bytes=16)


class sipHash128ReferenceKeyed(sipHash64Keyed):
    output_field = fields.FixedStringField(max_bytes=16)


class cityHash64(Func):
    output_field = fields.UInt64Field()


class intHash32(Func):
    arity = 1
    output_field = fields.UInt32Field()


class intHash64(Func):
    arity = 1
    output_field = fields.UInt64Field()


class SHA1(Func):
    arity = 1
    output_field = fields.FixedStringField(max_bytes=20)


class SHA224(Func):
    arity = 1
    output_field = fields.FixedStringField(max_bytes=28)


class SHA256(Func):
    arity = 1
    output_field = fields.FixedStringField(max_bytes=32)


class SHA512(Func):
    arity = 1
    output_field = fields.FixedStringField(max_bytes=64)


class SHA512_256(Func):
    arity = 1
    output_field = fields.FixedStringField(max_bytes=64)


class BLAKE3(Func):
    arity = 1
    output_field = fields.FixedStringField(max_bytes=32)


class URLHash(Func):
    output_field = fields.UInt64Field()

    def __init__(self, *expressions):
        if len(expressions) == 1:
            pass
        elif len(expressions) == 2:
            expressions = (expressions[0], Value(expressions[1]))
        else:
            raise TypeError(
                "'%s' takes 1 or 2 arguments (%s given)"
                % (
                    self.__class__.__name__,
                    len(expressions),
                )
            )
        super().__init__(*expressions)


class farmFingerprint64(Func):
    output_field = fields.UInt64Field()


class farmHash64(Func):
    output_field = fields.UInt64Field()



import sys

from clickhouse_driver.errors import ErrorCodes
from django.db.backends.base.creation import BaseDatabaseCreation
from django.db.backends.utils import strip_quotes
from django.conf import settings
from django.utils.module_loading import import_string


class DatabaseCreation(BaseDatabaseCreation):
    def _quote_name(self, name):
        return self.connection.ops.quote_name(name)

    def sql_table_creation_suffix(self):
        test_settings = self.connection.settings_dict["TEST"]
        cluster = test_settings.get("cluster")
        engine = test_settings.get("engine")

        parts = []
        if cluster:
            parts.append(f"ON CLUSTER {self.connection.ops.quote_name(cluster)}")
        if engine:
            parts.append(f"ENGINE = {engine}")
        return " ".join(parts)

    def _get_on_cluster(self):
        test_settings = self.connection.settings_dict["TEST"]
        cluster = test_settings.get("cluster")
        if cluster:
            return f"ON CLUSTER {self.connection.ops.quote_name(cluster)}"
        return ""

    def _database_exists(self, cursor, database_name):
        cursor.execute(
            "SELECT 1 FROM system.databases WHERE name = %s",
            [strip_quotes(database_name)],
        )
        return cursor.fetchone() is not None

    def create_test_db(
        self, verbosity=1, autoclobber=False, serialize=True, keepdb=False
    ):
        super().create_test_db(verbosity, autoclobber, serialize, keepdb)
        test_settings = self.connection.settings_dict["TEST"]
        if "fake_transaction" in test_settings:
            self.connection.fake_transaction = test_settings["fake_transaction"]
        self.mark_expected_failures_and_skips()

    def _create_test_db(self, verbosity, autoclobber, keepdb=False):
        """
        Internal implementation - create the test db tables.
        """
        if not self.connection.settings_dict["TEST"].get("managed", True):
            return
        test_database_name = self._get_test_db_name()
        test_db_params = {
            "dbname": self.connection.ops.quote_name(test_database_name),
            "suffix": self.sql_table_creation_suffix(),
        }
        # Create the test database and connect to it.
        with self._nodb_cursor() as cursor:
            try:
                self._execute_create_test_db(cursor, test_db_params, keepdb)
            except Exception as e:
                # if we want to keep the db, then no need to do any of the below,
                # just return and skip it all.
                if keepdb:
                    return test_database_name

                self.log("Got an error creating the test database: %s" % e)
                if not autoclobber:
                    confirm = input(
                        "Type 'yes' if you would like to try deleting the test "
                        "database '%s', or 'no' to cancel: " % test_database_name
                    )
                if autoclobber or confirm == "yes":
                    try:
                        if verbosity >= 1:
                            self.log(
                                "Destroying old test database for alias %s..."
                                % (
                                    self._get_database_display_str(
                                        verbosity, test_database_name
                                    ),
                                )
                            )
                        sql = "DROP DATABASE %(dbname)s" % test_db_params
                        on_cluster = self._get_on_cluster()
                        if on_cluster:
                            sql = f"{sql} {on_cluster} SYNC"
                        cursor.execute(sql)
                        self._execute_create_test_db(cursor, test_db_params, keepdb)
                    except Exception as e:
                        self.log("Got an error recreating the test database: %s" % e)
                        sys.exit(2)
                else:
                    self.log("Tests cancelled.")
                    sys.exit(1)

        return test_database_name

    def _execute_create_test_db(self, cursor, parameters, keepdb=False):
        try:
            if keepdb and self._database_exists(cursor, parameters["dbname"]):
                # If the database should be kept and it already exists, don't
                # try to create a new one.
                return
            super()._execute_create_test_db(cursor, parameters, keepdb)
        except Exception as e:
            if (
                not e.args
                or getattr(e.args[0], "code", "") != ErrorCodes.DATABASE_ALREADY_EXISTS
            ):
                # All errors except "database already exists" cancel tests.
                self.log("Got an error creating the test database: %s" % e)
                sys.exit(2)
            elif not keepdb:
                # If the database should be kept, ignore "database already
                # exists".
                raise

    def _destroy_test_db(self, test_database_name, verbosity):
        test_settings = self.connection.settings_dict["TEST"]
        if not test_settings.get("managed", True):
            return
        sql = "DROP DATABASE %s" % self.connection.ops.quote_name(test_database_name)
        on_cluster = self._get_on_cluster()
        if on_cluster:
            sql = f"{sql} {on_cluster} SYNC"
        with self._nodb_cursor() as cursor:
            cursor.execute(sql)

    def mark_expected_failures_and_skips(self):
        """
        Mark tests in Django's test suite which are expected failures on this
        database and test which should be skipped on this database.
        """
        # Only load unittest if we're actually testing.
        from unittest import expectedFailure, skip

        for test_name in self.connection.features.django_test_expected_failures:
            test_case_name, _, test_method_name = test_name.rpartition(".")
            test_app = test_name.split(".")[0]
            # Importing a test app that isn't installed raises RuntimeError.
            if test_app in settings.INSTALLED_APPS:
                test_case = import_string(test_case_name)
                test_method = getattr(test_case, test_method_name)
                setattr(test_case, test_method_name, expectedFailure(test_method))
        for reason, tests in self.connection.features.django_test_skips.items():
            for test_name in tests:
                test_case_name, _, test_method_name = test_name.rpartition(".")
                test_app = test_name.split(".")[0]
                # Importing a test app that isn't installed raises RuntimeError.
                if test_app in settings.INSTALLED_APPS:
                    test_case = import_string(test_case_name)
                    test_method = getattr(test_case, test_method_name)
                    setattr(test_case, test_method_name, skip(reason)(test_method))



from collections import namedtuple

from django.db import router
from django.db.models.sql import query, subqueries
from django.db.models.sql.constants import INNER
from django.db.models.sql.datastructures import BaseTable, Join
from django.db.models.sql.where import AND

from clickhouse_backend import compat

ExplainInfo = namedtuple("ExplainInfo", ("format", "type", "options"))


class Query(query.Query):
    def __init__(self, model, where=query.WhereNode, alias_cols=True):
        if compat.dj_ge4:
            super().__init__(model, alias_cols)
        else:
            super().__init__(model, where, alias_cols)
        self.setting_info = {}
        self.prewhere = query.WhereNode()

    def sql_with_params(self):
        """Choose the right db when database router is used."""
        return self.get_compiler(router.db_for_read(self.model)).as_sql()

    def clone(self):
        obj = super().clone()
        obj.setting_info = self.setting_info.copy()
        obj.prewhere = self.prewhere.clone()
        return obj

    def explain(self, using, format=None, type=None, **settings):
        q = self.clone()
        q.explain_info = ExplainInfo(format, type, settings)
        compiler = q.get_compiler(using=using)
        return "\n".join(compiler.explain_query())

    def add_prewhere(self, q_object):
        """
        refer add_q
        """
        existing_inner = {
            a for a in self.alias_map if self.alias_map[a].join_type == INNER
        }
        clause, _ = self._add_q(q_object, self.used_aliases)
        if clause:
            self.prewhere.add(clause, AND)
        self.demote_joins(existing_inner)

    if not compat.dj_ge42:

        @property
        def is_sliced(self):
            return self.low_mark != 0 or self.high_mark is not None

    def resolve_expression(self, query, *args, **kwargs):
        clone = self.clone()
        # Subqueries need to use a different set of aliases than the outer query.
        clone.bump_prefix(query)
        clone.subquery = True
        clone.where.resolve_expression(query, *args, **kwargs)
        clone.prewhere.resolve_expression(query, *args, **kwargs)
        # Resolve combined queries.
        if clone.combinator:
            clone.combined_queries = tuple(
                [
                    combined_query.resolve_expression(query, *args, **kwargs)
                    for combined_query in clone.combined_queries
                ]
            )
        for key, value in clone.annotations.items():
            resolved = value.resolve_expression(query, *args, **kwargs)
            if hasattr(resolved, "external_aliases"):
                resolved.external_aliases.update(clone.external_aliases)
            clone.annotations[key] = resolved
        # Outer query's aliases are considered external.
        for alias, table in query.alias_map.items():
            clone.external_aliases[alias] = (
                isinstance(table, Join)
                and table.join_field.related_model._meta.db_table != alias
            ) or (
                isinstance(table, BaseTable) and table.table_name != table.table_alias
            )
        return clone

    def change_aliases(self, change_map):
        """
        Change the aliases in change_map (which maps old-alias -> new-alias),
        relabelling any references to them in select columns and the where
        clause.
        """
        # If keys and values of change_map were to intersect, an alias might be
        # updated twice (e.g. T4 -> T5, T5 -> T6, so also T4 -> T6) depending
        # on their order in change_map.
        assert set(change_map).isdisjoint(change_map.values())

        # 1. Update references in "select" (normal columns plus aliases),
        # "group by" and "where".
        self.where.relabel_aliases(change_map)
        self.prewhere.relabel_aliases(change_map)
        if isinstance(self.group_by, tuple):
            self.group_by = tuple(
                [col.relabeled_clone(change_map) for col in self.group_by]
            )
        self.select = tuple([col.relabeled_clone(change_map) for col in self.select])
        self.annotations = self.annotations and {
            key: col.relabeled_clone(change_map)
            for key, col in self.annotations.items()
        }

        # 2. Rename the alias in the internal table/alias datastructures.
        for old_alias, new_alias in change_map.items():
            if old_alias not in self.alias_map:
                continue
            alias_data = self.alias_map[old_alias].relabeled_clone(change_map)
            self.alias_map[new_alias] = alias_data
            self.alias_refcount[new_alias] = self.alias_refcount[old_alias]
            del self.alias_refcount[old_alias]
            del self.alias_map[old_alias]

            table_aliases = self.table_map[alias_data.table_name]
            for pos, alias in enumerate(table_aliases):
                if alias == old_alias:
                    table_aliases[pos] = new_alias
                    break
        self.external_aliases = {
            # Table is aliased or it's being changed and thus is aliased.
            change_map.get(alias, alias): (aliased or alias in change_map)
            for alias, aliased in self.external_aliases.items()
        }


def clone_decorator(cls):
    old_clone = cls.clone

    def clone(self):
        obj = old_clone(self)
        if hasattr(obj, "setting_info"):
            obj.setting_info = self.setting_info.copy()
        return obj

    cls.clone = clone
    return cls


clone_decorator(subqueries.UpdateQuery)
clone_decorator(subqueries.DeleteQuery)



import warnings

from django.db.backends.base.schema import (
    BaseDatabaseSchemaEditor,
    _related_non_m2m_objects,
)
from django.db.backends.ddl_references import (
    Columns,
    Expressions,
    IndexName,
    Statement,
    Table,
)
from django.db.models.constraints import CheckConstraint, UniqueConstraint
from django.db.models.expressions import Col, ExpressionList, F
from django.db.models.indexes import IndexExpression

from clickhouse_backend import compat
from clickhouse_backend.driver.escape import escape_param


class ChColumns(Columns):
    def __str__(self):
        def col_str(column, idx):
            col = self.quote_name(column)
            try:
                suffix = self.col_suffixes[idx]
                if suffix:
                    col = "-{}".format(col)
            except IndexError:
                pass
            return col

        return ", ".join(
            col_str(column, idx) for idx, column in enumerate(self.columns)
        )


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    sql_create_table = "CREATE TABLE %(table)s %(on_cluster)s (%(definition)s) ENGINE = %(engine)s %(extra)s"
    sql_rename_table = "RENAME TABLE %(old_table)s TO %(new_table)s %(on_cluster)s"
    sql_delete_table = "DROP TABLE %(table)s %(on_cluster)s"

    sql_create_column = (
        "ALTER TABLE %(table)s %(on_cluster)s ADD COLUMN %(column)s %(definition)s"
    )
    sql_alter_column = "ALTER TABLE %(table)s %(on_cluster)s %(changes)s"
    sql_alter_column_type = "MODIFY COLUMN %(column)s %(type)s"
    sql_alter_column_null = sql_alter_column_type
    sql_alter_column_not_null = sql_alter_column_type
    sql_alter_column_default = "MODIFY COLUMN %(column)s DEFAULT %(default)s"
    sql_alter_column_no_default = "MODIFY COLUMN %(column)s REMOVE DEFAULT"
    sql_alter_column_no_default_null = sql_alter_column_no_default
    sql_delete_column = "ALTER TABLE %(table)s %(on_cluster)s DROP COLUMN %(column)s"
    sql_rename_column = "ALTER TABLE %(table)s %(on_cluster)s RENAME COLUMN %(old_column)s TO %(new_column)s"
    sql_update_with_default = (
        "ALTER TABLE %(table)s %(on_cluster)s UPDATE %(column)s = %(default)s "
        "WHERE %(column)s IS NULL SETTINGS mutations_sync=1"
    )

    sql_create_check = (
        "ALTER TABLE %(table)s %(on_cluster)s ADD CONSTRAINT %(name)s CHECK (%(check)s)"
    )
    sql_delete_check = "ALTER TABLE %(table)s %(on_cluster)s DROP CONSTRAINT %(name)s"

    sql_index = "INDEX %(name)s (%(columns)s) TYPE %(type)s GRANULARITY %(granularity)s"
    sql_create_index = "ALTER TABLE %(table)s %(on_cluster)s ADD " + sql_index
    sql_delete_index = "ALTER TABLE %(table)s %(on_cluster)s DROP INDEX %(name)s"

    sql_create_constraint = "ALTER TABLE %(table)s %(on_cluster)s ADD %(constraint)s"

    sql_alter_column_comment = (
        "ALTER TABLE %(table)s %(on_cluster)s COMMENT COLUMN %(column)s %(comment)s"
    )
    sql_alter_table_comment = (
        "ALTER TABLE %(table)s %(on_cluster)s MODIFY COMMENT %(comment)s"
    )

    def delete_model(self, model):
        """Delete a model from the database."""
        # Handle auto-created intermediary models
        for field in model._meta.local_many_to_many:
            if field.remote_field.through._meta.auto_created:
                self.delete_model(field.remote_field.through)

        # Delete the table
        self.execute(
            self.sql_delete_table
            % {
                "table": self.quote_name(model._meta.db_table),
                "on_cluster": self._get_on_cluster(model),
            }
        )
        # Remove all deferred statements referencing the deleted table.
        for sql in list(self.deferred_sql):
            if isinstance(sql, Statement) and sql.references_table(
                model._meta.db_table
            ):
                self.deferred_sql.remove(sql)

    def _column_check_name(self, field):
        return "_check_%s" % field.column

    def _column_check_sql(self, field):
        db_params = field.db_parameters(connection=self.connection)
        if db_params["check"]:
            return self._check_sql(
                name=self._column_check_name(field),
                check=self.sql_check_constraint % db_params,
            )

    def table_sql(self, model):
        # https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/mergetree/#table_engine-mergetree-creating-a-table
        # TODO: Support SAMPLE/TTL/SETTINGS
        """Take a model and return its table definition."""
        # Create column SQL, include constraint and index.
        column_sqls = []
        params = []
        constraints = []
        for field in model._meta.local_fields:
            definition, extra_params = self.column_sql(model, field)
            if definition is None:
                continue
            params.extend(extra_params)
            # Add the SQL to our big list.
            column_sqls.append(
                "%s %s"
                % (
                    self.quote_name(field.column),
                    definition,
                )
            )
            constraints.append(self._column_check_sql(field))
        for constraint in model._meta.constraints:
            if isinstance(constraint, CheckConstraint):
                constraints.append(constraint.constraint_sql(model, self))
        if any(isinstance(c, UniqueConstraint) for c in model._meta.constraints):
            warnings.warn("Clickhouse does not support unique constraint.")

        engine = self._get_engine(model)
        extra_parts = self._model_extra_sql(model, engine)
        sql = self.sql_create_table % {
            "table": self.quote_name(model._meta.db_table),
            "on_cluster": self._get_on_cluster(model),
            "definition": ", ".join(
                str(constraint)
                for constraint in (*column_sqls, *constraints)
                if constraint
            ),
            "engine": self._get_engine_expression(model, engine),
            "extra": " ".join(extra_parts),
        }
        return sql, params

    def _get_on_cluster(self, model, sync=False):
        cluster = getattr(model._meta, "cluster", None)
        if cluster:
            if sync:
                return f"ON CLUSTER {self.quote_name(cluster)} SYNC"
            return f"ON CLUSTER {self.quote_name(cluster)}"
        return ""

    def _get_engine(self, model):
        from clickhouse_backend.models import MergeTree

        return getattr(
            model._meta, "engine", MergeTree(order_by=model._meta.pk.attname)
        )

    def _get_engine_expression(self, model, engine):
        from clickhouse_backend.models.sql import Query

        compiler = Query(model, alias_cols=False).get_compiler(
            connection=self.connection
        )
        engine.source_expressions = tuple(
            Col("", model._meta.get_field(e.name)) if isinstance(e, F) else e
            for e in engine.source_expressions
        )
        return Expressions(model._meta.db_table, engine, compiler, self.quote_value)

    def column_sql(self, model, field, include_default=False):
        # https://clickhouse.com/docs/en/sql-reference/statements/create/table/#create-table-query
        # TODO: Support [MATERIALIZED|EPHEMERAL|ALIAS expr1] [compression_codec] [TTL expr1]
        """
        Take a field and return its column definition.
        The field must already have had set_attributes_from_name() called.
        """
        # Get the column's type and use that as the basis of the SQL
        db_params = field.db_parameters(connection=self.connection)
        sql = db_params["type"]
        params = []
        # Check for fields that aren't actually columns (e.g. M2M)
        if sql is None:
            return None, None
        if field.null and "Nullable" not in sql:  # Compatible with django fields.
            sql = f"Nullable({sql})"
        # Add database default.
        if compat.field_has_db_default(field):
            default_sql, default_params = self.db_default_sql(field)
            sql = f"{sql} DEFAULT {default_sql}"
            params.extend(default_params)
        elif include_default:
            default_value = self.effective_default(field)
            if default_value is not None:
                sql = f"{sql} DEFAULT {self._column_default_sql(field)}"
                params.append(default_value)
        if compat.field_db_comment(field):
            sql = f"{sql} COMMENT %s"
            params.append(field.db_comment)
        return sql, params

    def _model_indexes_sql(self, model):
        """
        Return a list of all index SQL statements for the specified model.
        field indexes and index_together are ignored, only Meta.indexes is considered.
        """
        if not model._meta.managed or model._meta.proxy or model._meta.swapped:
            return []
        output = []

        msg = (
            "Because index requires extra params, such as TYPE and GRANULARITY, "
            "so field level index=True and Meta level index_together is ignored. "
            "Refer to https://clickhouse.com/docs/en/engines/table-engines/"
            "mergetree-family/mergetree/#table_engine-mergetree-data_skipping-indexes"
        )
        if any(field.db_index for field in model._meta.local_fields) or getattr(
            model._meta, "index_together", None
        ):
            warnings.warn(msg)

        for index in model._meta.indexes:
            output.append(index.create_sql(model, self))

        from clickhouse_backend.models.engines import BaseMergeTree

        engine = self._get_engine(model)
        if output and not isinstance(engine, BaseMergeTree):
            raise ValueError(
                "Index manipulation is supported only for tables with "
                "*MergeTree engine (including replicated variants). Refer to "
                "https://clickhouse.com/docs/en/sql-reference/statements/alter/index."
            )
        return output

    def _get_expression(self, model, *expressions):
        if not expressions:
            return ""
        index_expressions = []
        for expression in expressions:
            index_expression = IndexExpression(expression)
            index_expression.set_wrapper_classes(self.connection)
            index_expressions.append(index_expression)
        from clickhouse_backend.models.sql import Query

        query = Query(model, alias_cols=False)
        expression_list = ExpressionList(*index_expressions).resolve_expression(query)
        compiler = query.get_compiler(connection=self.connection)
        return Expressions(
            model._meta.db_table, expression_list, compiler, self.quote_value
        )

    def _model_extra_sql(self, model, engine):
        from clickhouse_backend.models.engines import BaseMergeTree

        if isinstance(engine, BaseMergeTree):
            order_by = engine.order_by
            partition_by = engine.partition_by
            primary_key = engine.primary_key

            if order_by is not None:
                yield "ORDER BY (%s)" % self._get_expression(model, *order_by)
            if partition_by:
                yield "PARTITION BY (%s)" % self._get_expression(model, *partition_by)
            if primary_key is not None:
                yield "PRIMARY KEY (%s)" % self._get_expression(model, *primary_key)
        if engine.settings:
            result = []
            for setting, value in engine.settings.items():
                result.append(f"{setting}={self.quote_value(value)}")
            yield "SETTINGS %s" % ", ".join(result)

    def alter_db_table_comment(self, model, old_db_table_comment, new_db_table_comment):
        self.execute(
            self.sql_alter_table_comment
            % {
                "table": self.quote_name(model._meta.db_table),
                "comment": self.quote_value(new_db_table_comment or ""),
                "on_cluster": self._get_on_cluster(model),
            }
        )

    def add_field(self, model, field):
        """
        Create a field on a model. Usually involves adding a column, but may
        involve adding a table instead (for M2M fields).
        """
        # Special-case implicit M2M tables
        if field.many_to_many and field.remote_field.through._meta.auto_created:
            return self.create_model(field.remote_field.through)
        # Get the column's definition
        definition, params = self.column_sql(model, field, include_default=False)
        # It might not actually have a column behind it
        if definition is None:
            return

        check_sql = self._column_check_sql(field)
        if check_sql:
            self.deferred_sql.append(
                self.sql_create_constraint
                % {
                    "table": self.quote_name(model._meta.db_table),
                    "constraint": check_sql,
                    "on_cluster": self._get_on_cluster(model),
                }
            )

        # Build the SQL and run it
        sql = self.sql_create_column % {
            "table": self.quote_name(model._meta.db_table),
            "column": self.quote_name(field.column),
            "definition": definition,
            "on_cluster": self._get_on_cluster(model),
        }
        self.execute(sql, params)
        # Drop the default if we need to
        # (Django usually does not use in-database defaults)
        if (
            not compat.field_has_db_default(field)
            and self.effective_default(field) is not None
        ):
            # Update existing rows with default value
            sql_update_default = (
                "ALTER TABLE %(table)s %(on_cluster)s UPDATE %(column)s = %(default)s "
                "WHERE 1 SETTINGS mutations_sync=1"
            )
            from clickhouse_backend import models

            if not isinstance(self._get_engine(model), models.Distributed):
                self.execute(
                    sql_update_default
                    % {
                        "table": self.quote_name(model._meta.db_table),
                        "column": self.quote_name(field.column),
                        "default": "%s",
                        "on_cluster": self._get_on_cluster(model),
                    },
                    [self.effective_default(field)],
                )

    def remove_field(self, model, field):
        """
        Remove a field from a model. Usually involves deleting a column,
        but for M2Ms may involve deleting a table.
        """
        # Special-case implicit M2M tables
        if field.many_to_many and field.remote_field.through._meta.auto_created:
            return self.delete_model(field.remote_field.through)
        # It might not actually have a column behind it
        db_params = field.db_parameters(connection=self.connection)
        if db_params["type"] is None:
            return
        # Delete the column
        sql = self.sql_delete_column % {
            "table": self.quote_name(model._meta.db_table),
            "column": self.quote_name(field.column),
            "on_cluster": self._get_on_cluster(model),
        }
        self.execute(sql)
        if db_params["check"]:
            constraint_name = self._column_check_name(field)
            self.execute(self._delete_check_sql(model, constraint_name))
        # Reset connection if required
        if self.connection.features.connection_persists_old_columns:
            self.connection.close()
        # Remove all deferred statements referencing the deleted column.
        for sql in list(self.deferred_sql):
            if isinstance(sql, Statement) and sql.references_column(
                model._meta.db_table, field.column
            ):
                self.deferred_sql.remove(sql)

    def quote_value(self, value):
        if isinstance(value, str):
            value = value.replace("%", "%%")
        return escape_param(value, {})

    def _field_indexes_sql(self, model, field):
        return []

    def _field_data_type(self, field):
        if field.is_relation:
            return field.rel_db_type(self.connection)
        return self.connection.data_types.get(
            field.get_internal_type(),
            field.db_type(self.connection),
        )

    def _field_base_data_types(self, field):
        # Yield base data types for array fields.
        if field.base_field.get_internal_type() == "ArrayField":
            yield from self._field_base_data_types(field.base_field)
        else:
            yield self._field_data_type(field.base_field)

    def _field_should_be_altered(self, old_field, new_field):
        if (not (old_field.concrete or old_field.many_to_many)) and (
            not (new_field.concrete or new_field.many_to_many)
        ):
            return False
        _, old_path, old_args, old_kwargs = old_field.deconstruct()
        _, new_path, new_args, new_kwargs = new_field.deconstruct()
        # Don't alter when:
        # - changing only a field name
        # - changing an attribute that doesn't affect the schema
        # - adding only a db_column and the column name is not changed
        non_database_attrs = (
            "verbose_name",
            # field primary_key is an internal concept of django,
            # clickhouse MergeTree primary_key is a different concept.
            "primary_key",
            # Clickhouse dont have unique constraint
            "unique",
            "blank",
            # Clickhouse don't support inline index, use meta index instead
            "db_index",
            "editable",
            "serialize",
            "unique_for_date",
            "unique_for_month",
            "unique_for_year",
            "help_text",
            "db_column",
            "db_tablespace",
            "auto_created",
            "validators",
            "error_messages",
            "on_delete",
            "related_name",
            "related_query_name",
            "db_collation",
            "limit_choices_to",
            "size",
        )

        from clickhouse_backend.models import EnumField

        if not isinstance(old_field, EnumField):
            old_kwargs.pop("choices", None)
        if not isinstance(new_field, EnumField):
            new_kwargs.pop("choices", None)

        for attr in non_database_attrs:
            old_kwargs.pop(attr, None)
            new_kwargs.pop(attr, None)

        if (
            not new_field.many_to_many
            and old_field.remote_field
            and new_field.remote_field
            and old_field.remote_field.model._meta.db_table
            == new_field.remote_field.model._meta.db_table
        ):
            old_kwargs.pop("to", None)
            new_kwargs.pop("to", None)
        # db_default can take many form but result in the same SQL.
        if (
            compat.dj_ge5
            and old_kwargs.get("db_default")
            and new_kwargs.get("db_default")
            and self.db_default_sql(old_field) == self.db_default_sql(new_field)
        ):
            old_kwargs.pop("db_default")
            new_kwargs.pop("db_default")
        if (
            old_field.concrete
            and new_field.concrete
            and (self.quote_name(old_field.column) != self.quote_name(new_field.column))
        ):
            return True
        if (
            old_field.many_to_many
            and new_field.many_to_many
            and old_field.name != new_field.name
        ):
            return True
        return (old_path, old_args, old_kwargs) != (new_path, new_args, new_kwargs)

    def _get_column_type(self, field):
        column_type = field.db_type(connection=self.connection)
        if field.null and "Nullable" not in column_type:
            column_type = f"Nullable({column_type})"
        return column_type

    def _alter_column_null_sql(self, model, old_field, new_field):
        """
        Hook to specialize column null alteration.

        Return a (sql, params) fragment to set a column to null or non-null
        as required by new_field, or None if no changes are required.
        """
        # Compatible for django fields.
        new_type = self._get_column_type(new_field)
        return (
            self.sql_alter_column_type
            % {
                "column": self.quote_name(new_field.column),
                "type": new_type,
            },
            [],
        )

    def _alter_field(
        self,
        model,
        old_field,
        new_field,
        old_type,
        new_type,
        old_db_params,
        new_db_params,
        strict=False,
    ):
        old_type = self._get_column_type(old_field)
        new_type = self._get_column_type(new_field)
        # Change check constraints?
        if old_db_params["check"] and (
            old_db_params["check"] != new_db_params["check"]
            or old_field.column != new_field.column
        ):
            constraint_name = self._column_check_name(old_field)
            self.execute(self._delete_check_sql(model, constraint_name))
        # Have they renamed the column?
        if old_field.column != new_field.column:
            self.execute(
                self.sql_rename_column
                % {
                    "table": self.quote_name(model._meta.db_table),
                    "old_column": self.quote_name(old_field.column),
                    "new_column": self.quote_name(new_field.column),
                    "type": new_type,
                    "on_cluster": self._get_on_cluster(model),
                }
            )
            # Rename all references to the renamed column.
            for sql in self.deferred_sql:
                if isinstance(sql, Statement):
                    sql.rename_column_references(
                        model._meta.db_table, old_field.column, new_field.column
                    )
        # Next, start accumulating actions to do
        actions = []
        post_actions = []
        # Only if we have a default and there is a change from NULL to NOT NULL
        four_way_default_alteration = (
            new_field.has_default() or compat.field_has_db_default(new_field)
        ) and (old_field.null and not new_field.null)
        # Type or comment change?
        if old_type != new_type or compat.field_db_comment(
            old_field
        ) != compat.field_db_comment(new_field):
            # Should not alter nullable before null values updated.
            if four_way_default_alteration:
                new_field.null = True
                new_type_backup = new_type
                new_type = self._get_column_type(new_field)
            fragment, other_actions = self._alter_column_type_sql(
                model, old_field, new_field, new_type
            )
            if four_way_default_alteration:
                new_field.null = False
                new_type = new_type_backup
            actions.append(fragment)
            post_actions.extend(other_actions)

        if compat.field_has_db_default(new_field):
            if (
                not compat.field_has_db_default(old_field)
                or new_field.db_default != old_field.db_default
            ):
                actions.append(
                    self._alter_column_database_default_sql(model, old_field, new_field)
                )
        elif compat.field_has_db_default(old_field):
            actions.append(
                self._alter_column_database_default_sql(
                    model, old_field, new_field, drop=True
                )
            )
        # When changing a column NULL constraint to NOT NULL with a given
        # default value, we need to perform 4 steps:
        #  1. Add a default for new incoming writes
        #  2. Update existing NULL rows with new default
        #  3. Replace NULL constraint with NOT NULL
        #  4. Drop the default again.
        # Default change?
        needs_database_default = False
        if (
            old_field.null
            and not new_field.null
            and not compat.field_has_db_default(new_field)
        ):
            old_default = self.effective_default(old_field)
            new_default = self.effective_default(new_field)
            if old_default != new_default and new_default is not None:
                needs_database_default = True
                actions.append(
                    self._alter_column_default_sql(model, old_field, new_field)
                )
        if actions:
            sql, params = tuple(zip(*actions))
            sql, params = (", ".join(sql), sum(params, []))
            self.execute(
                self.sql_alter_column
                % {
                    "table": self.quote_name(model._meta.db_table),
                    "changes": sql,
                    "on_cluster": self._get_on_cluster(model),
                },
                params,
            )
        if four_way_default_alteration:
            from clickhouse_backend.models import Distributed

            if not isinstance(self._get_engine(model), Distributed):
                if not compat.field_has_db_default(new_field):
                    default_sql = "%s"
                    params = [new_default]
                else:
                    default_sql, params = self.db_default_sql(new_field)
                # Update existing rows with default value
                self.execute(
                    self.sql_update_with_default
                    % {
                        "table": self.quote_name(model._meta.db_table),
                        "column": self.quote_name(new_field.column),
                        "default": default_sql,
                        "on_cluster": self._get_on_cluster(model),
                    },
                    params,
                )
            # Since we didn't run a NOT NULL change before we need to do it now
            sql, params = self._alter_column_null_sql(model, old_field, new_field)
            self.execute(
                self.sql_alter_column
                % {
                    "table": self.quote_name(model._meta.db_table),
                    "changes": sql,
                    "on_cluster": self._get_on_cluster(model),
                },
                params,
            )
        if post_actions:
            for sql, params in post_actions:
                self.execute(sql, params)
        # Type alteration on primary key? Then we need to alter the column
        # referring to us.
        drop_foreign_keys = (old_field.primary_key and new_field.primary_key) and (
            old_type != new_type
        )
        rels_to_update = []
        if drop_foreign_keys:
            rels_to_update.extend(_related_non_m2m_objects(old_field, new_field))
        # Changed to become primary key?
        if self._field_became_primary_key(old_field, new_field):
            # Update all referencing columns
            rels_to_update.extend(_related_non_m2m_objects(old_field, new_field))
        # Handle our type alters on the other end of rels from the PK stuff above
        for old_rel, new_rel in rels_to_update:
            rel_db_params = new_rel.field.db_parameters(connection=self.connection)
            rel_type = rel_db_params["type"]
            if (
                new_rel.field.null and "Nullable" not in rel_type
            ):  # Compatible with django fields.
                rel_type = "Nullable(%s)" % rel_type
            fragment, other_actions = self._alter_column_type_sql(
                new_rel.related_model, old_rel.field, new_rel.field, rel_type
            )
            self.execute(
                self.sql_alter_column
                % {
                    "table": self.quote_name(new_rel.related_model._meta.db_table),
                    "changes": fragment[0],
                    "on_cluster": self._get_on_cluster(model),
                },
                fragment[1],
            )
            for sql, params in other_actions:
                self.execute(sql, params)
        # Does it have check constraints we need to add?
        if new_db_params["check"] and (
            old_db_params["check"] != new_db_params["check"]
            or old_field.column != new_field.column
        ):
            check_sql = self._column_check_sql(new_field)
            if check_sql:
                self.execute(
                    self.sql_create_constraint
                    % {
                        "table": self.quote_name(model._meta.db_table),
                        "constraint": check_sql,
                        "on_cluster": self._get_on_cluster(model),
                    }
                )
        # Drop the default if we need to
        # (Django usually does not use in-database defaults)
        if needs_database_default:
            changes_sql, params = self._alter_column_default_sql(
                model, old_field, new_field, drop=True
            )
            sql = self.sql_alter_column % {
                "table": self.quote_name(model._meta.db_table),
                "changes": changes_sql,
                "on_cluster": self._get_on_cluster(model),
            }
            self.execute(sql, params)

    def _alter_column_type_sql(self, model, old_field, new_field, new_type):
        if compat.dj_ge42:
            return super()._alter_column_type_sql(
                model, old_field, new_field, new_type, "", ""
            )
        else:
            return super()._alter_column_type_sql(model, old_field, new_field, new_type)

    def _alter_column_comment_sql(self, model, new_field, new_type, new_db_comment):
        return (
            self.sql_alter_column_comment
            % {
                "table": self.quote_name(model._meta.db_table),
                "column": self.quote_name(new_field.column),
                "comment": self._comment_sql(new_db_comment),
                "on_cluster": self._get_on_cluster(model),
            },
            [],
        )

    def _create_index_sql(
        self,
        model,
        *,
        fields=None,
        name=None,
        sql=None,
        suffix="",
        col_suffixes=None,
        type=None,
        granularity=None,
        expressions=None,
        inline=False,
        **kwargs,
    ):
        """
        Return the SQL statement to create the index for one or several fields
        or expressions. `sql` can be specified if the syntax differs from the
        standard (GIS indexes, ...).
        """
        fields = fields or []
        expressions = expressions or []
        from clickhouse_backend.models.sql import Query

        compiler = Query(model, alias_cols=False).get_compiler(
            connection=self.connection,
        )
        columns = [field.column for field in fields]
        sql_create_index = sql or (self.sql_index if inline else self.sql_create_index)
        table = model._meta.db_table

        def create_index_name(*args, **kwargs):
            nonlocal name
            if name is None:
                name = self._create_index_name(*args, **kwargs)
            return self.quote_name(name)

        return Statement(
            sql_create_index,
            table=Table(table, self.quote_name),
            on_cluster=self._get_on_cluster(model),
            name=IndexName(table, columns, suffix, create_index_name),
            columns=(
                Columns(table, columns, self.quote_name, col_suffixes)
                if columns
                else Expressions(table, expressions, compiler, self.quote_value)
            ),
            type=Expressions(table, type, compiler, self.quote_value),
            granularity=granularity,
        )

    def _delete_index_sql(self, model, name, sql=None):
        return Statement(
            sql or self.sql_delete_index,
            table=Table(model._meta.db_table, self.quote_name),
            name=self.quote_name(name),
            on_cluster=self._get_on_cluster(model),
        )

    def alter_unique_together(self, model, old_unique_together, new_unique_together):
        """for django or other third party app, ignore unique constraint.
        User defined app should never use UniqueConstraint"""
        pass

    def alter_db_table(self, model, old_db_table, new_db_table):
        """Rename the table a model points to."""
        if old_db_table == new_db_table:
            return
        self.execute(
            self.sql_rename_table
            % {
                "old_table": self.quote_name(old_db_table),
                "new_table": self.quote_name(new_db_table),
                "on_cluster": self._get_on_cluster(model),
            }
        )
        # Rename all references to the old table name.
        for sql in self.deferred_sql:
            if isinstance(sql, Statement):
                sql.rename_table_references(old_db_table, new_db_table)

    def _create_check_sql(self, model, name, check):
        return Statement(
            self.sql_create_check,
            table=Table(model._meta.db_table, self.quote_name),
            name=self.quote_name(name),
            check=check,
            on_cluster=self._get_on_cluster(model),
        )

    def _delete_check_sql(self, model, name):
        return self._delete_constraint_sql(self.sql_delete_check, model, name)

    def _delete_constraint_sql(self, template, model, name):
        return Statement(
            template,
            table=Table(model._meta.db_table, self.quote_name),
            name=self.quote_name(name),
            on_cluster=self._get_on_cluster(model),
        )

    def add_constraint(self, model, constraint):
        """Add a constraint to a model."""
        from clickhouse_backend import models

        if isinstance(self._get_engine(model), models.Distributed):
            raise TypeError("Distributed table engine does not support constraints.")
        sql = constraint.create_sql(model, self)
        if sql:
            # Constraint.create_sql returns interpolated SQL which makes
            # params=None a necessity to avoid escaping attempts on execution.
            self.execute(sql, params=None)

    def remove_constraint(self, model, constraint):
        """Remove a constraint from a model."""
        from clickhouse_backend import models

        if isinstance(self._get_engine(model), models.Distributed):
            raise TypeError("Distributed table engine does not support constraints.")
        sql = constraint.remove_sql(model, self)
        if sql:
            self.execute(sql)



import re
import time

from clickhouse_driver import client

from .escape import escape_params

insert_pattern = re.compile(r"^\s*insert\s+into.+?values\s*;?$", flags=re.IGNORECASE)


class Client(client.Client):
    def __init__(self, *args, **kwargs):
        # https://clickhouse.com/docs/en/sql-reference/data-types/datetime/#usage-remarks
        # The clickhouse-client applies the server time zone by default
        # if a time zone isn’t explicitly set when initializing the data type.
        # To use the client time zone, run clickhouse-client with the --use_client_time_zone parameter.
        settings = kwargs.pop("settings", None) or {}
        settings.setdefault("use_client_time_zone", True)
        kwargs.update(settings=settings)
        super().__init__(*args, **kwargs)

    def substitute_params(self, query, params, context):
        escaped = escape_params(params, context)
        return query % escaped

    def execute(
        self,
        query,
        params=None,
        with_column_types=False,
        external_tables=None,
        query_id=None,
        settings=None,
        types_check=False,
        columnar=False,
    ):
        """Support dict params for INSERT queries."""
        start_time = time.time()

        with self.disconnect_on_error(query, settings):
            is_insert = insert_pattern.match(query)

            if is_insert:
                rv = self.process_insert_query(
                    query,
                    params,
                    external_tables=external_tables,
                    query_id=query_id,
                    types_check=types_check,
                    columnar=columnar,
                )
            else:
                rv = self.process_ordinary_query(
                    query,
                    params=params,
                    with_column_types=with_column_types,
                    external_tables=external_tables,
                    query_id=query_id,
                    types_check=types_check,
                    columnar=columnar,
                )
            self.last_query.store_elapsed(time.time() - start_time)
            return rv



from django.db.models import F, Q, query
from django.db.models.functions import Trunc

from clickhouse_backend.models import fields, sql


class QuerySet(query.QuerySet):
    def explain(self, *, format=None, type=None, **settings):
        """
        Runs an EXPLAIN on the SQL query this QuerySet would perform, and
        returns the results.
        https://clickhouse.com/docs/en/sql-reference/statements/explain/
        """
        return self.query.explain(using=self.db, format=format, type=type, **settings)

    def settings(self, **kwargs):
        clone = self._chain()
        if isinstance(clone.query, sql.Query):
            clone.query.setting_info.update(kwargs)
        return clone

    def prewhere(self, *args, **kwargs):
        """
        Return a new QuerySet instance with the args ANDed to the existing
        prewhere set.
        """
        self._not_support_combined_queries("prewhere")
        if (args or kwargs) and self.query.is_sliced:
            raise TypeError("Cannot prewhere a query once a slice has been taken.")
        clone = self._chain()
        clone._query.add_prewhere(Q(*args, **kwargs))
        return clone

    def datetimes(self, field_name, kind, order="ASC", tzinfo=None):
        """
        Return a list of datetime objects representing all available
        datetimes for the given field_name, scoped to 'kind'.
        """
        if kind not in ("year", "month", "week", "day", "hour", "minute", "second"):
            raise ValueError(
                "'kind' must be one of 'year', 'month', 'week', 'day', "
                "'hour', 'minute', or 'second'."
            )
        if order not in ("ASC", "DESC"):
            raise ValueError("'order' must be either 'ASC' or 'DESC'.")

        if kind in ("year", "month", "week"):
            output_field = fields.DateField()
        else:
            output_field = fields.DateTimeField()
        return (
            self.annotate(
                datetimefield=Trunc(
                    field_name,
                    kind,
                    output_field=output_field,
                    tzinfo=tzinfo,
                ),
                plain_field=F(field_name),
            )
            .values_list("datetimefield", flat=True)
            .distinct()
            .filter(plain_field__isnull=False)
            .order_by(("-" if order == "DESC" else "") + "datetimefield")
        )



from django.db import InterfaceError
from django.db.backends.base.features import BaseDatabaseFeatures
from django.utils.functional import cached_property


class DatabaseFeatures(BaseDatabaseFeatures):
    # TODO: figure out compatible clickhouse versions.
    minimum_database_version = None
    # Use this class attribute control whether using fake transaction.
    # Fake transaction is used in test, prevent other database such as postgresql
    # from flush at the end of each testcase. Only use this feature when you are
    # aware of the effect in TransactionTestCase.
    fake_transaction = False
    # Clickhouse do support Geo Data Types, but they are based on GeoJSON instead GIS.
    # https://clickhouse.com/docs/en/sql-reference/data-types/geo/
    # gis_enabled = False

    # There is no dedicated LOB type in clickhouse_backend, LOB is stored as string type.
    # https://clickhouse.com/docs/en/sql-reference/data-types/string/
    # allows_group_by_lob = True

    # There is no unique constraint in clickhouse_backend.
    # supports_nullable_unique_constraints = True
    # supports_partially_nullable_unique_constraints = True
    # supports_deferrable_unique_constraints = False

    # Clickhouse only supports limited transaction.
    # https://clickhouse.com/docs/en/sql-reference/ansi/
    # https://github.com/ClickHouse/ClickHouse/issues/32513
    # https://clickhouse.com/docs/en/guides/developer/transactional
    @cached_property
    def uses_savepoints(self):
        return self.fake_transaction

    can_release_savepoints = False

    # Is there a true datatype for uuid?
    has_native_uuid_field = True

    # Clickhouse use re2 syntax which does not support backreference.
    # https://clickhouse.com/docs/en/sql-reference/functions/string-search-functions#matchhaystack-pattern
    # https://github.com/google/re2/wiki/Syntax
    supports_regex_backreferencing = False

    # Can date/datetime lookups be performed using a string?
    supports_date_lookup_using_string = True

    # Confirm support for introspected foreign keys
    # Every database can do this reliably, except MySQL,
    # which can't do it for MyISAM tables
    can_introspect_foreign_keys = False

    # Map fields which some backends may not be able to differentiate to the
    # field it's introspected as.

    introspected_field_types = {
        "AutoField": "Int64Field",
        "BigAutoField": "Int64Field",
        "BigIntegerField": "Int64Field",
        "BinaryField": "StringField",
        "BooleanField": "BoolField",
        "CharField": "FixedStringField",
        # "DurationField": "DurationField",
        "GenericIPAddressField": "IPv6Field",
        "IntegerField": "Int32Field",
        "PositiveBigIntegerField": "UInt64Field",
        "PositiveIntegerField": "UInt32Field",
        "PositiveSmallIntegerField": "UInt16Field",
        "SmallAutoField": "Int64Field",
        "SmallIntegerField": "Int16Field",
        # "TimeField": "TimeField",
    }

    # https://clickhouse.com/docs/en/sql-reference/statements/alter/index/
    # Index manipulation is supported only for tables with *MergeTree* engine (including replicated variants).
    # Not btree index, don't support ordering.
    supports_index_column_ordering = False

    # Does the backend support introspection of materialized views?
    can_introspect_materialized_views = True

    # Support for the DISTINCT ON clause
    can_distinct_on_fields = True

    # Does the backend prevent running SQL queries in broken transactions?
    atomic_transactions = False

    # Can we issue more than one ALTER COLUMN clause in an ALTER TABLE?
    supports_combined_alters = True

    # Does it support foreign keys?
    supports_foreign_keys = False

    # Does it support CHECK constraints?
    supports_column_check_constraints = False
    supports_table_check_constraints = True
    # Does the backend support introspection of CHECK constraints?
    can_introspect_check_constraints = True

    # What kind of error does the backend throw when accessing closed cursor?
    closed_cursor_error_class = InterfaceError

    # https://clickhouse.com/docs/en/sql-reference/statements/insert-into/#constraints
    supports_ignore_conflicts = False

    # Does the backend support partial indexes (CREATE INDEX ... WHERE ...)?
    supports_partial_indexes = False

    # Does the backend support JSONField?
    supports_json_field = True
    # Can the backend introspect a JSONField?
    can_introspect_json_field = True
    # Does the backend support primitives in JSONField?
    supports_primitives_in_json_field = False
    # Is there a true datatype for JSON?
    has_native_json_field = True
    # Does the backend use PostgreSQL-style JSON operators like '->'?
    has_json_operators = False
    # Does the backend support __contains and __contained_by lookups for
    # a JSONField?
    supports_json_field_contains = False
    # Does value__d__contains={'f': 'g'} (without a list around the dict) match
    # {'d': [{'f': 'g'}]}?
    json_key_contains_list_matching_requires_list = False
    # Does the backend support JSONObject() database function?
    has_json_object_function = False

    # Does the backend support column collations?
    supports_collation_on_charfield = False
    supports_collation_on_textfield = False
    # Does the backend support non-deterministic collations?
    supports_non_deterministic_collations = False

    # Does the backend support column and table comments?
    supports_comments = True
    # Does the backend support column comments in ADD COLUMN statements?
    supports_comments_inline = True

    # SQL template override for tests.aggregation.tests.NowUTC
    test_now_utc_template = "now64()"

    # Does the database support SQL 2003 FILTER (WHERE ...) in aggregate
    # expressions?
    supports_aggregate_filter_clause = True

    # Does the backend support window expressions (expression OVER (...))?
    supports_over_clause = True
    supports_frame_range_fixed_distance = True
    only_supports_unbounded_with_preceding_and_following = False
    insert_test_table_with_defaults = (
        "INSERT INTO {} VALUES (DEFAULT, DEFAULT, DEFAULT)"
    )

    @cached_property
    def supports_transactions(self):
        return self.fake_transaction

    @cached_property
    def django_test_skips(self):
        skips = {}
        version = self.connection.get_database_version()
        if version >= (25, 11):
            skips.update(
                {
                    "ClickHouse 25.11 remove deprecated Object('json') type.": {
                        "clickhouse_fields.test_jsonfield.JsonFieldTests.test_query",
                        "expressions_window.tests.WindowFunctionTests.test_key_transform",
                    }
                }
            )
        if version < (25, 1):
            skips.update(
                {
                    "ClickHouse 25.1 add generateSerialID function.": {
                        "clickhouse_functions.test_other.OtherTests.test_generateSerialID",
                    }
                }
            )
        return skips

    @cached_property
    def django_test_expected_failures(self):
        # DB::Exception: Cannot convert column 'height' from nullable type Nullable(UInt32) to non-nullable type UInt32.
        # Please specify `DEFAULT` expression in ALTER MODIFY COLUMN statement.
        # A bug in ClickHouse 25.11, they force a `DEFAULT` expression when alter column type from nullable to non-nullable.
        if self.connection.get_database_version() >= (25, 11):
            return {
                "schema.tests.SchemaTests.test_alter_null_to_not_null_keeping_default",
                "schema.tests.SchemaTests.test_alter",
            }
        return set()



from django.core import checks
from django.core.exceptions import ImproperlyConfigured


class FieldMixin:
    """All clickhouse field should inherit this mixin.

    1. Remove unsupported arguments: unique, db_index, unique_for_date,
       unique_for_month, unique_for_year, db_tablespace, db_collation.
    2. Return shortened name in deconstruct method.

    low_cardinality argument is added separately in every specific field that support LowCardinality.
    If added in this mixin, then PyCharm will not supply argument hints.
    """

    nullable_allowed = True

    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_nullable(),
        ]

    def _check_nullable(self):
        if self.null and not self.nullable_allowed:
            return [
                checks.Error(
                    "Nullable is not supported by %s." % self.__class__.__name__,
                    obj=self,
                )
            ]
        return []

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        for key in [
            "unique",
            "db_index",
            "unique_for_date",
            "unique_for_month",
            "unique_for_year",
            "db_tablespace",
            "db_collation",
        ]:
            try:
                del kwargs[key]
            except KeyError:
                pass

        if getattr(self, "low_cardinality", False):
            kwargs["low_cardinality"] = self.low_cardinality
        if path.startswith("clickhouse_backend.models.fields"):
            path = path.replace(
                "clickhouse_backend.models.fields", "clickhouse_backend.models"
            )
        return name, path, args, kwargs

    def _nested_type(self, value):
        if value is not None:
            # LowCardinality and Nullable sequence matters, reversal will cause DB::Exception:
            # Nested type LowCardinality(Int8) cannot be inside Nullable type. (ILLEGAL_TYPE_OF_ARGUMENT)
            if self.null:
                value = "Nullable(%s)" % value
            if getattr(self, "low_cardinality", False):
                value = "LowCardinality(%s)" % value
        return value

    def _check_backend(self, connection):
        if connection.vendor != "clickhouse":
            raise ImproperlyConfigured(
                "%s must only be used with django clickhouse backend."
                % self.__class__.__name__
            )

    def db_type(self, connection):
        self._check_backend(connection)
        return self._nested_type(super().db_type(connection))



from django.db.models import functions

__all__ = [
    "patch_functions",
    "patch_random",
]


def patch_functions():
    patch_now()
    patch_random()


def patch_now():
    def as_clickhouse(self, compiler, connection, **extra_context):
        return functions.Now.as_sql(
            self, compiler, connection, template="now64()", **extra_context
        )

    functions.Now.as_clickhouse = as_clickhouse


def patch_random():
    def as_clickhouse(self, compiler, connection, **extra_context):
        return functions.Random.as_sql(
            self, compiler, connection, function="rand64", **extra_context
        )

    functions.Random.as_clickhouse = as_clickhouse



import django
from django.db import models

dj3 = (3,) <= django.VERSION < (4,)
dj4 = (4,) <= django.VERSION < (5,)
dj_ge4 = django.VERSION >= (4,)
dj_ge41 = django.VERSION >= (4, 1)
dj_ge42 = django.VERSION >= (4, 2)
dj_ge5 = django.VERSION >= (5,)
dj_ge51 = django.VERSION >= (5, 1)
dj_ge52 = django.VERSION >= (5, 2)
dj_ge6 = django.VERSION >= (6,)


def db_table_comment(model: models.Model) -> str:
    """return a model's database table comment.

    https://docs.djangoproject.com/en/4.2/releases/4.2/#comments-on-columns-and-tables
    """
    return dj_ge42 and model._meta.db_table_comment or ""


def field_db_comment(field: models.Field) -> str:
    """return a field's database column comment.

    https://docs.djangoproject.com/en/4.2/releases/4.2/#comments-on-columns-and-tables
    """
    return dj_ge42 and field.db_comment or ""


def field_has_db_default(field: models.Field) -> bool:
    """check if a field has database level default value.

    https://docs.djangoproject.com/en/5.0/releases/5.0/#database-computed-default-values
    """
    if dj_ge52:
        return field.has_db_default()
    if dj_ge5:
        return field.db_default is not models.NOT_PROVIDED
    return False



from collections.abc import Iterable

from django.db.models import Func, Value

__all__ = [
    "Distributed",
    "Engine",
    "BaseMergeTree",
    "MergeTree",
    "ReplacingMergeTree",
    "GraphiteMergeTree",
    "CollapsingMergeTree",
    "VersionedCollapsingMergeTree",
    "SummingMergeTree",
    "AggregatingMergeTree",
    "ReplicatedMergeTree",
    "ReplicatedReplacingMergeTree",
    "ReplicatedGraphiteMergeTree",
    "ReplicatedCollapsingMergeTree",
    "ReplicatedVersionedCollapsingMergeTree",
    "ReplicatedSummingMergeTree",
    "ReplicatedAggregatingMergeTree",
]


def _check_positive(value, name):
    if not isinstance(value, int) and value <= 0:
        raise ValueError(f"{name} must be positive integer.")
    return value


def _check_not_negative(value, name):
    if not isinstance(value, int) and value < 0:
        raise ValueError(f"{name} must not be negative.")
    return value


def _check_bool(value, name):
    if value not in (0, 1):
        raise ValueError(f"{name} must be one of (0, 1, True, False)")
    return int(value)


def _check_str(value, name):
    if not isinstance(value, str):
        raise ValueError(f"{name} must be string")
    return value


def value_if_string(value):
    if isinstance(value, str):
        return Value(value)
    return value


class Engine(Func):
    max_arity = None  # The max number of arguments the function accepts.
    setting_types = {}

    def __init__(self, *expressions, **settings):
        if self.max_arity is not None and len(expressions) > self.max_arity:
            raise TypeError(
                "'%s' takes at most %s %s (%s given)"
                % (
                    self.__class__.__name__,
                    self.max_arity,
                    "argument" if self.max_arity == 1 else "arguments",
                    len(expressions),
                )
            )

        normalized_settings = {}
        for setting, value in settings.items():
            for validate, keys in self.setting_types.items():
                if setting in keys:
                    normalized_settings[setting] = validate(value, setting)
                    break
            else:
                raise TypeError(f"{setting} is not a valid setting.")
        self.settings = normalized_settings
        super().__init__(*expressions)

    @property
    def function(self):
        return self.__class__.__name__

    def deconstruct(self):
        path, args, kwargs = super().deconstruct()
        if path.startswith("clickhouse_backend.models.engines"):
            path = path.replace(
                "clickhouse_backend.models.engines", "clickhouse_backend.models"
            )
        return path, args, kwargs


class BaseMergeTree(Engine):
    # https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/mergetree#settings
    setting_types = {
        _check_positive: [
            "index_granularity",
            "min_index_granularity_bytes",
            "merge_max_block_size",
            "min_bytes_for_wide_part",
            "min_rows_for_wide_part",
            "max_parts_in_total",
            "max_compress_block_size",
            "min_compress_block_size",
            "max_partitions_to_read",
        ],
        _check_not_negative: [
            "index_granularity_bytes",
            "min_merge_bytes_to_use_direct_io",
            "merge_with_ttl_timeout",
            "merge_with_recompression_ttl_timeout",
            "try_fetch_recompressed_part_timeout",
        ],
        _check_bool: [
            "enable_mixed_granularity_parts",
            "use_minimalistic_part_header_in_zookeeper",
            "write_final_mark",
        ],
        _check_str: ["storage_policy"],
    }

    def __init__(
        self,
        *expressions,
        order_by=None,
        partition_by=None,
        primary_key=None,
        **settings,
    ):
        assert order_by is not None or primary_key is not None, (
            "At least one of order_by or primary_key must be provided"
        )
        self.order_by = order_by
        self.primary_key = primary_key
        self.partition_by = partition_by

        for key in ["order_by", "primary_key", "partition_by"]:
            value = getattr(self, key)
            if value is not None:
                if isinstance(value, str) or not isinstance(value, Iterable):
                    value = (value,)
                    setattr(self, key, value)
                elif not isinstance(value, tuple):
                    value = tuple(value)
                    setattr(self, key, value)
                if any(i is None for i in value):
                    raise ValueError(f"None is not allowed in {key}")

        # https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/mergetree#choosing-a-primary-key-that-differs-from-the-sorting-key
        # primary key expression tuple must be a prefix of the sorting key expression tuple.
        if (
            self.order_by is not None
            and self.primary_key is not None
            and self.order_by[: len(self.primary_key)] != self.primary_key
        ):
            raise ValueError("primary_key must be a prefix of order_by")

        super().__init__(*expressions, **settings)


class MergeTree(BaseMergeTree):
    arity = 0


class ReplacingMergeTree(BaseMergeTree):
    max_arity = 2


class SummingMergeTree(BaseMergeTree):
    pass


class AggregatingMergeTree(BaseMergeTree):
    arity = 0


class CollapsingMergeTree(BaseMergeTree):
    arity = 1


class VersionedCollapsingMergeTree(BaseMergeTree):
    arity = 2


class GraphiteMergeTree(BaseMergeTree):
    def __init__(self, config_section, **extra):
        super().__init__(value_if_string(config_section), **extra)


class ReplicatedMixin:
    def __init__(self, *replicated_parameters, other_parameters=(), **extra):
        """
        https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/replication#replicatedmergetree-parameters
        https://github.com/ClickHouse/ClickHouse/issues/8675
        https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/replication#creating-replicated-tables

        replicated_parameters: zoo_path and replica_name. Their default values can be specified in the server
            configuration file, in this case, they both can be omitted.
        other_parameters: Parameters of an engine which is used for creating the replicated version,
            for example, version in ReplacingMergeTree
        """
        if replicated_parameters:
            if len(replicated_parameters) != 2:
                raise TypeError(
                    "'ReplicatedMergeTree' takes 0 or 2 arguments (%s given)"
                    % len(replicated_parameters)
                )
            replicated_parameters = map(value_if_string, replicated_parameters)
        super().__init__(*other_parameters, **extra)
        self.source_expressions = (*replicated_parameters, *self.source_expressions)


class ReplicatedMergeTree(ReplicatedMixin, MergeTree):
    pass


class ReplicatedReplacingMergeTree(ReplicatedMixin, ReplacingMergeTree):
    pass


class ReplicatedSummingMergeTree(ReplicatedMixin, SummingMergeTree):
    pass


class ReplicatedAggregatingMergeTree(ReplicatedMixin, AggregatingMergeTree):
    pass


class ReplicatedCollapsingMergeTree(ReplicatedMixin, CollapsingMergeTree):
    pass


class ReplicatedVersionedCollapsingMergeTree(
    ReplicatedMixin, VersionedCollapsingMergeTree
):
    pass


class ReplicatedGraphiteMergeTree(ReplicatedMixin, GraphiteMergeTree):
    pass


class Distributed(Engine):
    # https://clickhouse.com/docs/en/engines/table-engines/special/distributed#distributed-settings
    setting_types = {
        _check_positive: ["monitor_sleep_time_ms", "monitor_max_sleep_time_ms"],
        _check_not_negative: [
            "bytes_to_throw_insert",
            "bytes_to_delay_insert",
            "max_delay_to_insert",
        ],
        _check_bool: [
            "fsync_after_insert",
            "fsync_directories",
            "monitor_batch_inserts",
            "monitor_split_batch_on_failure",
        ],
    }

    def __init__(self, *expressions, **settings):
        if len(expressions) < 3:
            raise TypeError(
                "'%s' takes at least 3 arguments (cluster, database, table)"
                % self.__class__.__name__
            )
        if len(expressions) > 5:
            raise TypeError(
                "'%s' takes at most 5 arguments (cluster, database, table[, sharding_key[, policy_name]])"
                % self.__class__.__name__
            )

        self.cluster, self.database, self.table = expressions[:3]
        if len(expressions) == 5:
            expressions = (
                *map(value_if_string, expressions[:3]),
                expressions[3],
                value_if_string(expressions[4]),
            )
        else:
            expressions = (*map(value_if_string, expressions[:3]), *expressions[3:])
        super().__init__(*expressions, **settings)



from clickhouse_driver.util.compat import get_localzone_name_compat
from django.conf import settings
from django.utils import timezone


def get_timezone():
    """Guess what timezone the user is in.

    Just remember, no matter USE_TZ or not,
    ClickHouse always store UNIX timestamp (which is in UTC timezone).

    Results of lookups turning datetime to date are relevant to timezone.
    We should always pass timezone parameter to ClickHouse functions like
    toDate, toStartOfMonth.
    """
    if settings.USE_TZ:
        try:
            return timezone.get_current_timezone_name()
        except TypeError:
            # When TIME_ZONE = None
            pass
    return get_localzone_name_compat()
